"""
    harness
    =======

    Simple test harness to simplify testing asynchronous code.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

import aiohttp
import asyncio
import collections
from collections import abc
import contextlib
import copy
import dataclasses
import datetime
import enum
import functools
import inspect
import json
import math
import requests
import string
import struct
import typing
import unittest
import warnings
import websockets

from xpxchain import client
from xpxchain import models
from xpxchain import util
from . import aitertools
from . import responses

# HELPERS


def get_argnames(func):
    """Get the argument names from a function."""

    # Get all positional arguments in __init__, including named
    # and named optional arguments. `co_varnames` stores all argument
    # names (including local variable names) in order, starting with
    # function arguments, so only grab `co_argcount` varnames.
    code = getattr(func, '__func__', func).__code__
    argcount = code.co_argcount
    return code.co_varnames[:argcount]


def setattr_default(obj, name, value):
    """Setdefault except for object attributes."""

    if name not in obj.__dict__:
        setattr(obj, name, value)


def add_tests(cls, tests, attrs, prefix):
    """Add tests to a class from a list of names and attrs passed to the class."""

    for test in tests:
        if attrs.get(test, True):
            name = f'test_{test}'
            func = globals()[f'{prefix}_test_{test}']
            setattr_default(cls, name, func)


def add_format_tests(cls, tests, attrs, prefix):
    """Add tests which require extra format data to be provided."""

    for test in tests:
        if test in attrs:
            name = f'test_{test}'
            func = globals()[f'{prefix}_test_{test}']
            setattr_default(cls, name, func)


def is_exception(obj):
    """Check if an object is an exception."""
    return inspect.isclass(obj) and issubclass(obj, Exception)


# TEST CASE


class TestCase(unittest.TestCase):
    """Asynchronous-aware test case suite."""

    def __init__(self, methodName='runTest', loop=None):
        self.loop = loop or asyncio.get_event_loop()
        self._cache = {}
        super().__init__(methodName=methodName)

    def coroutine_function_decorator(self, func):
        def wrapper(*args, **kw):
            return self.loop.run_until_complete(func(*args, **kw))
        return wrapper

    def __getattribute__(self, item):
        attr = object.__getattribute__(self, item)
        if asyncio.iscoroutinefunction(attr):
            if item not in self._cache:
                self._cache[item] = self.coroutine_function_decorator(attr)
            return self._cache[item]
        return attr


# ENUM TEST CASE

ENUM_TESTS = [
    'description',
    'value',
    'repr',
    'str',
    'eq',
    'ord',
]


def enum_test_description(self):
    """Test the enum description method."""

    for enumeration, description in zip(self.enums, self.descriptions):
        self.assertTrue(enumeration.description().startswith(description))


def enum_test_value(self):
    """Test the enum value property."""

    for enumeration, value in zip(self.enums, self.values):
        self.assertEqual(value, enumeration.value)


def enum_test_repr(self):
    """Test the enum __repr__ method."""

    clsname = self.type.__name__
    for enumeration in self.enums:
        expected = f'<{clsname}.{enumeration.name}: {repr(enumeration.value)}>'
        self.assertEqual(repr(enumeration), expected)


def enum_test_str(self):
    """Test the enum __str__ method."""

    clsname = self.type.__name__
    for enumeration in self.enums:
        self.assertEqual(str(enumeration), f'{clsname}.{enumeration.name}')


def enum_test_eq(self):
    """Test the enum __eq__ method."""

    enums = list(set(self.enums))
    for index, enumeration in enumerate(enums):
        self.assertEqual(enumeration, enums[index])
        for other in enums[index + 1:]:
            self.assertNotEqual(enumeration, other)


def enum_test_ord(self):
    """Test the enum __lt__ method."""

    enums = sorted(set(self.enums))
    for index, enumeration in enumerate(enums):
        self.assertLessEqual(enumeration, enums[index])
        for other in enums[index + 1:]:
            self.assertLess(enumeration, other)


def enum_test_dto(self):
    """Test the conversion to and from DTO."""

    for enumeration, dto in zip(self.enums, self.dto):
        self.assertEqual(dto, enumeration.to_dto())


def enum_test_catbuffer(self):
    """Test the conversion to and from catbuffer."""

    for enumeration, catbuffer in zip(self.enums, self.catbuffer):
        self.assertEqual(catbuffer, enumeration.to_catbuffer())


def enum_setup_fn(attrs):
    """Generate the setup function for enums."""

    assert len(attrs['enums']) == len(attrs['descriptions'])
    assert len(attrs['enums']) == len(attrs['values'])
    assert 'dto' not in attrs or len(attrs['enums']) == len(attrs['dto'])
    assert 'catbuffer' not in attrs or len(attrs['enums']) == len(attrs['catbuffer'])
    for test in attrs.get('custom', ()):
        inputs = test.get('inputs') or attrs['enums']
        results = test.get('results') or attrs['enums']
        assert len(inputs) == len(results)

    def func(self):
        self.type = attrs['type']
        self.enums = attrs['enums']
        self.values = attrs['values']
        self.descriptions = attrs['descriptions']
        self.dto = attrs.get('dto')
        self.catbuffer = attrs.get('catbuffer')
        self.extras = attrs.get('extras', {})

    return func


def enum_add_custom_tests(cls, tests):
    """Add custom tests, which provide a name, callback, and expected results."""

    def decorator(callback, inputs, results):
        def func(self):
            tests = zip(inputs or self.enums, results or self.enums)
            for in_value, out_value in tests:
                if is_exception(out_value):
                    with self.assertRaises(out_value):
                        callback(self, in_value)
                else:
                    self.assertEqual(out_value, callback(self, in_value))
        return func

    for test in tests:
        callback = test['callback']
        inputs = test.get('inputs')
        results = test.get('results')
        setattr_default(cls, test['name'], decorator(callback, inputs, results))


def enum_test_case(attrs):
    """
    Decorator that auto-generates the test suite from model data.

    Auto-generates and binds a model to the class, and then defines
    tests automatically from the bound data.

    Available tests are:
        test_description
        test_value
        test_repr
        test_str
        test_eq
        test_ord
    """

    def decorator(cls):
        setattr_default(cls, 'setUp', enum_setup_fn(attrs))
        add_tests(cls, ENUM_TESTS, attrs, prefix='enum')
        add_format_tests(cls, ['catbuffer', 'dto'], attrs, prefix='enum')
        enum_add_custom_tests(cls, attrs.get('custom', ()))
        return cls

    return decorator


# MODEL TEST CASE

MODEL_TESTS = [
    'init',
    'frozen',
    'slots',
    'eq',
    'repr',
    'str',
    'copy',
    'deepcopy',
    'replace',
    'asdict',
    'astuple',
    'fields',
]


def model_test_init(self):
    """Test if the model fields are properly set from the initializer."""

    if dataclasses.is_dataclass(self.type):
        for name, value in self.data.items():
            self.assertEqual(getattr(self.model, name), value)
    elif issubclass(self.type, abc.Sequence):
        self.assertEqual(tuple(self.data), tuple(self.model))
    elif issubclass(self.type, abc.Mapping):
        self.assertEqual(dict(self.data), dict(self.model))
    else:
        raise NotImplementedError


def model_test_frozen(self):
    """Test the model's attributes are frozen."""

    if dataclasses.is_dataclass(self.type):
        fields = [i.name for i in dataclasses.fields(self.model)]
        for field in fields:
            value = getattr(self.model, field)
            with self.assertRaises(dataclasses.FrozenInstanceError):
                setattr(self.model, field, value)
    elif issubclass(self.type, abc.Sequence):
        for index, value in enumerate(self.model):
            with self.assertRaises(TypeError):
                self.model[index] = value
    elif issubclass(self.type, abc.Mapping):
        for key, value in self.model.items():
            with self.assertRaises(TypeError):
                self.model[key] = value
    else:
        raise NotImplementedError


def model_test_slots(self):
    """Test if the model has slots, and therefore is efficiently stored."""

    with self.assertRaises((TypeError, AttributeError)):
        self.model.__dict__


def next_byte_power(value):
    """Calculate the next power of 2 from a value."""

    char_bit = 8
    byte_length = int(math.ceil(value.bit_length() / char_bit))
    return 2 ** (char_bit * byte_length)


def bitwise_not(value):
    """Calculate bitwise not of value as an unsigned integer."""
    # We want to get 1 - next power of 2, so we can get the 0xFF bit pattern.
    # We can then use xor to invert all the bits in the input.
    # We want to use value + 1, since if the value is, say, 256, it requires
    # 9 bits of storage, not 8.
    max_int = next_byte_power(value + 1) - 1
    return max_int ^ value


def model_test_eq(self):
    """Test the model __eq__ method."""

    def permute_sequence(data):
        return type(data)([permute(i) for i in data])

    def permute_mapping(data):
        return type(data)([(k, permute(v)) for k, v in data.items()])

    def permute(value):
        # In bytes and strings, the first 2 values may be used as a
        # sentinel. We want to keep those the same. 1 for bytes,
        # 2 for hex.
        if hasattr(value, '_permute_'):
            # Specialized method for testing.
            return value._permute_(permute)
        elif isinstance(value, enum.Enum):
            return value
        elif isinstance(value, datetime.datetime):
            timestamp = value.replace(tzinfo=datetime.timezone.utc).timestamp()
            negated = bitwise_not(int(timestamp))
            utc = datetime.datetime.fromtimestamp(negated, datetime.timezone.utc)
            return utc.replace(tzinfo=None)
        elif isinstance(value, int):
            return bitwise_not(value)
        elif isinstance(value, (bytes, bytearray)):
            return value[:1] + value[-2::-1]
        elif isinstance(value, str):
            return value[:2] + value[-3::-1]
        elif isinstance(value, abc.Sequence):
            return permute_sequence(value)
        elif isinstance(value, abc.Mapping):
            return permute_mapping(value)
        elif dataclasses.is_dataclass(value):
            fields = get_argnames(value.__init__)[1:]
            asdict = {i: getattr(value, i) for i in fields}
            return type(value)(**permute_mapping(asdict))
        return value

    m1 = copy.copy(self.model)
    m2 = copy.copy(self.model)
    m3 = permute(self.model)

    self.assertTrue(m1 == m1)
    self.assertTrue(m1 == m2)
    self.assertFalse(m1 == m3)
    self.assertTrue(m2 == m2)
    self.assertFalse(m2 == m3)
    self.assertTrue(m3 == m3)


def model_test_repr(self):
    """Test the model __repr__ method."""

    name = type(self.model).__name__
    if dataclasses.is_dataclass(self.type):
        fields = [i.name for i in dataclasses.fields(self.model)]
        values = [getattr(self.model, i) for i in fields]
        tupletype = collections.namedtuple(name, fields)
        self.assertEqual(repr(tupletype(*values)), repr(self.model))
    elif issubclass(self.type, abc.Sequence):
        self.assertEqual(repr(self.model), f'{name}({repr(list(self.model))})')
    elif issubclass(self.type, abc.Mapping):
        self.assertEqual(repr(self.model), f'{name}({repr(dict(self.model))})')
    else:
        raise NotImplementedError


def model_test_str(self):
    """Test the model __str__ method."""
    self.assertEqual(str(self.model), repr(self.model))


def model_test_copy(self):
    """Test the model __copy__ method."""
    self.assertEqual(self.model, copy.copy(self.model))


def model_test_deepcopy(self):
    """Test the model __deepcopy__ method."""
    self.assertEqual(self.model, copy.deepcopy(self.model))


def model_test_replace(self):
    """Test the model replace method."""
    self.assertEqual(self.model, self.model.replace())


def model_test_asdict(self):
    """Test the model asdict method."""

    asdict = self.model.asdict(recurse=False)
    if dataclasses.is_dataclass(self.type):
        fields = [i.name for i in dataclasses.fields(self.model)]
        self.assertIsInstance(asdict, dict)
        self.assertEqual(len(asdict), len(fields))
        for field in fields:
            self.assertEqual(asdict[field], getattr(self.model, field))
    elif issubclass(self.type, abc.Mapping):
        for key, value in self.model.items():
            self.assertEqual(asdict[key], value)
    else:
        raise NotImplementedError


def model_test_astuple(self):
    """Test the model astuple method."""

    astuple = self.model.astuple(recurse=False)
    if dataclasses.is_dataclass(self.type):
        fields = [i.name for i in dataclasses.fields(self.model)]
        self.assertIsInstance(astuple, tuple)
        self.assertEqual(len(astuple), len(fields))
        for index, field in enumerate(fields):
            self.assertEqual(astuple[index], getattr(self.model, field))
    elif issubclass(self.type, abc.Mapping):
        self.assertEqual(astuple, tuple(self.model.values()))
    elif issubclass(self.type, abc.Sequence):
        self.assertEqual(astuple, tuple(self.model))
    else:
        raise NotImplementedError


def model_test_fields(self):
    """Test the model fields method."""

    fields = self.model.fields()
    self.maxDiff = 2048
    self.assertIsInstance(fields, tuple)
    self.assertEqual(fields, dataclasses.fields(self.model))


def model_test_dto(self):
    """Test the conversion to and from DTO."""

    nt = self.network_type
    self.maxDiff = 2048
    self.assertEqual(self.model.to_dto(nt), self.dto)
    self.assertEqual(self.model, self.type.create_from_dto(self.dto, nt))


def model_test_catbuffer(self):
    """Test the conversion to and from catbuffer."""

    nt = self.network_type
    encoded = util.encode_hex(self.catbuffer)
    decoded = util.decode_hex(self.catbuffer)
    self.maxDiff = 2048
    self.assertEqual(util.hexlify(self.model.to_catbuffer(nt, fee_strategy=util.FeeCalculationStrategy.ZERO)), encoded)
    self.assertEqual(self.model, self.type.create_from_catbuffer(decoded, nt))


def load_model(model_type, data):
    """Generate the model from the provided data."""

    if issubclass(model_type, (abc.Sequence, abc.Mapping)):
        return model_type(data)
    return model_type(**data)


def model_setup_fn(attrs):
    """Generate the setup function for models."""

    model = load_model(attrs['type'], attrs['data'])

    def func(self):
        self.model = model
        self.type = attrs['type']
        self.data = attrs['data']
        self.network_type = attrs['network_type']
        self.dto = attrs.get('dto')
        self.catbuffer = attrs.get('catbuffer')
        self.extras = attrs.get('extras', {})

    return func


def model_test_case(attrs):
    """
    Decorator that auto-generates the test suite from model data.

    Auto-generates and binds a model to the class, and then defines
    tests automatically from the bound data.

    Available tests are:
        test_init
        test_frozen
        test_slots
        test_eq
        test_repr
        test_str
        test_copy
        test_deepcopy
        test_replace
        test_asdict
        test_astuple
        test_fields
    """

    def decorator(cls):
        setattr_default(cls, 'setUp', model_setup_fn(attrs))
        add_tests(cls, MODEL_TESTS, attrs, prefix='model')
        add_format_tests(cls, ['catbuffer', 'dto'], attrs, prefix='model')
        return cls

    return decorator


# TRANSACTION TEST CASE

TRANSACTION_TESTS = MODEL_TESTS + [
    'to_aggregate',
    'sign_with',
]

transaction_test_init = model_test_init
transaction_test_frozen = model_test_frozen
transaction_test_slots = model_test_slots
transaction_test_eq = model_test_eq
transaction_test_repr = model_test_repr
transaction_test_str = model_test_str
transaction_test_copy = model_test_copy
transaction_test_deepcopy = model_test_deepcopy
transaction_test_replace = model_test_replace
transaction_test_asdict = model_test_asdict
transaction_test_astuple = model_test_astuple
transaction_test_fields = model_test_fields
transaction_test_dto = model_test_dto
transaction_test_catbuffer = model_test_catbuffer


def transaction_test_to_aggregate(self):
    """Test embedding the transaction with a signer."""

    # Get the signer
    private_key = self.extras['private_key']
    signer = models.Account.create_from_private_key(private_key, self.network_type)

    # Convert to inner transaction and serialize.
    inner = self.model.to_aggregate(signer.public_account)
    catbuffer = inner.to_catbuffer(fee_strategy=self.extras['fee_strategy'])
    self.maxDiff = 2048
    self.assertEqual(util.hexlify(catbuffer), self.extras['embedded'])

    with self.assertRaises(TypeError):
        inner.__dict__


def transaction_test_sign_with(self):
    """Test transaction signing with a signer."""

    # Get the signer
    private_key = self.extras['private_key']
    signer = models.Account.create_from_private_key(private_key, self.network_type)

    # Sign transaction and check signed data.
    signed_transaction = self.model.sign_with(signer, self.extras['gen_hash'], fee_strategy=self.extras['fee_strategy'])
    self.maxDiff = 2048
    self.assertEqual(signed_transaction.payload, self.extras['signed']['payload'])
    self.assertEqual(signed_transaction.hash, self.extras['signed']['hash'])
    self.assertEqual(signed_transaction.signer, signer.public_key)
    self.assertEqual(signed_transaction.type, self.model.type)
    self.assertEqual(signed_transaction.network_type, self.network_type)


def transaction_test_case(attrs):
    """
    Decorator that auto-generates the test suite from transaction data.

    Auto-generates and binds a model to the class, and then defines
    tests automatically from the bound data. Supports all tests from
    model data, including a few more.

    Available tests are:
        test_to_aggregate
        test_sign_with
    """

    def decorator(cls):
        setattr_default(cls, 'setUp', model_setup_fn(attrs))
        add_tests(cls, TRANSACTION_TESTS, attrs, prefix='transaction')
        add_format_tests(cls, ['catbuffer', 'dto'], attrs, prefix='transaction')
        return cls

    return decorator


# HTTP TEST CASE


def make_request(http, method, params):
    """Make HTTP request from test data."""

    callback = getattr(http, method)
    return callback(*params)


def http_test(cls, sync_client, async_client, test):
    """Add a custom test for a mocked HTTP response."""

    @async_test(sync_client, async_client)
    async def func(self, data, await_cb, with_cb):
        async with with_cb(data(responses.ENDPOINT)) as http:
            request = make_request(http, test['method'], test['params'])
            response = await await_cb(request)
            for validator in test['validation']:
                self.assertEqual(*validator(response))

    func.__name__ = test['name']
    setattr_default(cls, test['name'], func)


def http_test_case(attrs):
    """Decorator that auto-generates HTTP test cases from attrs."""

    sync_client, async_client = attrs['clients']

    def decorator(cls):
        for test in attrs['tests']:
            http_test(cls, sync_client, async_client, test)
        return cls

    return decorator


# MOCKED HTTP TEST CASE


def mocked_http_test(cls, network_type, sync_client, async_client, test):
    """Add a custom test for a mocked HTTP response."""

    sync_data = (sync_client, requests)
    async_data = (async_client, aiohttp)
    network_response = responses.NETWORK_TYPE[network_type.name]
    method = test['method']
    params = test['params']
    error = test.get('error')

    @async_test(sync_data, async_data)
    async def func(self, data, await_cb, with_cb):
        async with with_cb(data[0](responses.ENDPOINT)) as http:
            # Set the network type, so we can make the actual request.
            with data[1].default_response(200, **network_response):
                await await_cb(http.network_type)

            # Make the actual request and validate the response.
            with data[1].default_response(200, **test['response']):
                if error is not None:
                    with self.assertRaises(error):
                        await await_cb(make_request(http, method, params))
                else:
                    response = await await_cb(make_request(http, method, params))
                    for validator in test.get('validation', ()):
                        self.assertEqual(*validator(response))

    func.__name__ = test['name']
    setattr_default(cls, test['name'], func)


def mocked_http_test_case(attrs):
    """Decorator that auto-generates mocked HTTP test cases from attrs."""

    sync_client, async_client = attrs['clients']
    network_type = attrs['network_type']

    def decorator(cls):
        for test in attrs['tests']:
            mocked_http_test(cls, network_type, sync_client, async_client, test)
        return cls

    return decorator


# LISTENER TEST


def listener_test(cls, test):
    """Add a custom test for a listener response."""

    async def func(self):
        async with client.Listener(f'{responses.ENDPOINT}/ws') as listener:
            # Subscribe to the desired channels.
            for subscription in test['subscriptions']:
                await getattr(listener, subscription)()

            # Iterate over all the messages.
            count = len(test['validation'])
            messages = [i async for i in aitertools.aslice(listener, count)]
            for message, validators in zip(messages, test['validation']):
                for validator in validators:
                    self.assertEqual(*validator(message))

    func.__name__ = test['name']
    setattr_default(cls, test['name'], func)


def listener_test_case(attrs):
    """Decorator that auto-generates listener test cases from attrs."""

    def decorator(cls):
        for test in attrs['tests']:
            listener_test(cls, test)
        return cls

    return decorator


# MOCKED LISTENER TEST


def mocked_listener_test(cls, test):
    """Add a custom test for a mocked listener response."""

    async def func(self):
        async with client.Listener(f'{responses.ENDPOINT}/ws') as listener:
            # Get the UID for the websockets client..
            with websockets.default_response([json.dumps({'uid': test['uid']})]):
                self.assertEqual(await listener.uid, test['uid'])

            # Subscribe to the desired channels.
            for subscription in test['subscriptions']:
                await getattr(listener, subscription)()

            # Iterate over all the messages.
            with websockets.default_response(test['response']):
                count = len(test['validation'])
                messages = [i async for i in aitertools.aslice(listener, count)]
                for message, validators in zip(messages, test['validation']):
                    for validator in validators:
                        self.assertEqual(*validator(message))

    func.__name__ = test['name']
    setattr_default(cls, test['name'], func)


def mocked_listener_test_case(attrs):
    """Decorator that auto-generates mocked listener test cases from attrs."""

    def decorator(cls):
        for test in attrs['tests']:
            mocked_listener_test(cls, test)
        return cls

    return decorator


# ASYNC TEST


def async_test(sync_data, async_data):
    """
    Generate synchronous and asynchronous tests from a single function.

    :param sync_data: Data to be passed to the synchronous test case.
    :param async_data: Data to be passed to the asynchronous test case.
    """

    def decorator(f):
        async def sync_cb(x):
            return x

        async def async_cb(x):
            return await x

        @contextlib.asynccontextmanager
        async def sync_with(x):
            with x as y:
                yield y

        @contextlib.asynccontextmanager
        async def async_with(x):
            async with x as y:
                yield y

        @functools.wraps(f)
        async def wrapped(self):
            await f(self, sync_data, sync_cb, sync_with)
            await f(self, async_data, async_cb, async_with)

        return wrapped

    return decorator

# IGNORE WARNING TEST


def ignore_warnings_test(test):
    """Wrap a function body in a warnings.catch_warnings() block."""

    def wrapper(self, *args, **kwds):
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            test(self, *args, **kwds)

    return wrapper


# RANDOM

with contextlib.suppress(ImportError):
    # Check we can import all our dependencies and generate the decorators.
    from collections import deque
    import logging
    import os.path
    import random
    import rstr
    import sys

    # CONFIG
    DEFAULT_CALLS = 20
    DEFAULT_MIN_STRLEN = 5
    DEFAULT_MAX_STRLEN = 40
    DEFAULT_MIN_SEQLEN = 1
    DEFAULT_MAX_SEQLEN = 10
    DEFAULT_MIN_MAPLEN = 1
    DEFAULT_MAX_MAPLEN = 10

    # TYPE HINTS
    # Define type hints so we can generate random data within the expected
    # bounds.
    U8 = typing.TypeVar('U8', bound=int)
    U16 = typing.TypeVar('U16', bound=int)
    U32 = typing.TypeVar('U32', bound=int)
    U64 = typing.TypeVar('U64', bound=int)
    U128 = typing.TypeVar('U128', bound=int)
    I8 = typing.TypeVar('I8', bound=int)
    I16 = typing.TypeVar('I16', bound=int)
    I32 = typing.TypeVar('I32', bound=int)
    I64 = typing.TypeVar('I64', bound=int)
    I128 = typing.TypeVar('I128', bound=int)
    F32 = typing.TypeVar('F32', bound=float)
    F64 = typing.TypeVar('F64', bound=float)

    class BytesPattern(bytes):
        """Indicates you want to match against a bytes regular expression."""

    class StrPattern(str):
        """Indicates you want to match against a str regular expression."""

    def boolgen():
        """Define a new generator for boolean values."""

        def generator() -> bool:
            return random.choice([True, False])

        return generator

    def intgen(min_value: int = -sys.maxsize, max_value: int = sys.maxsize):
        """Define a new generator for integral values (inclusive)."""

        def generator() -> int:
            return random.randint(min_value, max_value)

        return generator

    def structgen(bits: int, format: str):
        """Generate data from random bytes."""

        assert struct.calcsize(f'<{format}') == bits // 8
        nbytes = bits // 8

        def generator():
            data = int.to_bytes(random.getrandbits(bits), nbytes, 'little')
            return struct.unpack(f'<{format}', data)[0]

        return generator

    def floatgen(
        min_magnitude: float = 1e-7,
        max_magnitude: float = 1e10,
        positive: bool = False,
        **kwds
    ):
        """
        Define a new generator for floating-point values.
        Chance of 0 scales with the size of the range.
        """

        next_pos = boolgen()
        zero_bits = kwds.get('zero_bits', int(math.log2(max_magnitude - min_magnitude)))

        def generator() -> float:
            # Floats have a symmetric range for both positive and negative
            # values due to dedicated sign bit in IEEE754.
            if random.getrandbits(zero_bits) == 0:
                return 0.0
            elif positive or next_pos():
                return random.uniform(min_magnitude, max_magnitude)
            return random.uniform(-min_magnitude, -max_magnitude)

        return generator

    def regen(pattern: str):
        """Define a new generator based off a regular expression."""

        def generator() -> str:
            return rstr.xeger(pattern)

        return generator

    def lettergen(letters: str = string.printable, **kwds):
        """Define new letter-based string generator."""

        fixed_length = kwds.get('fixed_length')
        if fixed_length is not None:
            min_length = fixed_length
            max_length = fixed_length
        else:
            min_length = kwds.get('min_length', DEFAULT_MIN_STRLEN)
            max_length = kwds.get('max_length', DEFAULT_MAX_STRLEN)

        def generator() -> str:
            length = random.randint(min_length, max_length)
            return ''.join([random.choice(letters) for i in range(length)])

        return generator

    def strgen(**kwds):
        """Define a new generator for str."""

        if 'pattern' in kwds:
            return regen(kwds.pop('pattern'), **kwds)
        return lettergen(**kwds)

    def bytesgen(letters: bytes = bytes(range(256)), **kwds):
        """Define a new generator for bytes."""

        fixed_length = kwds.get('fixed_length')
        if fixed_length is not None:
            min_length = fixed_length
            max_length = fixed_length
        else:
            min_length = kwds.get('min_length', DEFAULT_MIN_STRLEN)
            max_length = kwds.get('max_length', DEFAULT_MAX_STRLEN)

        def generator() -> bytes:
            length = random.randint(min_length, max_length)
            return bytes([random.choice(letters) for i in range(length)])

        return generator

    def enumgen(type: typing.Type[enum.Enum]):
        """Randomly generate variants of an enumeration."""

        members = list(type.__members__.values())

        def generator() -> enum.Enum:
            return random.choice(members)

        return generator

    _T = typing.TypeVar('_T')
    _U = typing.TypeVar('_U')

    def _generator(type: typing.Type, kwds: typing.Optional[dict] = None):
        """Convert a type annotation and keyword arguments to a new generator."""

        # First, try an enumeration.
        if inspect.isclass(type) and issubclass(type, enum.Enum):
            return enumgen(type)

        # Next, try pre-defined generators.
        with contextlib.suppress(KeyError):
            return GENERATORS[type]

        # Try a sequence or mapping type. This will fail for TypeVars and Unions,
        # raising a KeyError by design. Ignore that.
        kwds = kwds or {}
        with contextlib.suppress(AttributeError):
            origin = type.__origin__
            args = type.__args__
            return PROTO_GENERATORS[origin](*args, **kwds)

        # Go to our fallback.
        return PROTO_GENERATORS[type](**kwds)

    def _seqgen(value_type: typing.Type[_T], **kwds):
        """Generate sequence type with defined value type."""

        # Process our arguments for the value type generator.
        valuegen = _generator(value_type, kwds.get('value'))

        # Process our arguments for the sequence generator.
        fixed_length = kwds.get('fixed_length')
        if fixed_length is not None:
            min_length = fixed_length
            max_length = fixed_length
        else:
            min_length = kwds.get('min_length', DEFAULT_MIN_SEQLEN)
            max_length = kwds.get('max_length', DEFAULT_MAX_SEQLEN)

        def generator() -> typing.Sequence[_T]:
            length = random.randint(min_length, max_length)
            return [valuegen() for _ in range(length)]

        return generator

    def _mapgen(key_type: typing.Type[_T], value_type: typing.Type[_U], **kwds):
        """Generate mapping type with defined key/value types."""

        # Process our arguments for the key and value type generators.
        keygen = _generator(key_type, kwds.get('key'))
        valuegen = _generator(value_type, kwds.get('value'))

        # Process our arguments for the sequence generator.
        fixed_length = kwds.get('fixed_length')
        if fixed_length is not None:
            min_length = fixed_length
            max_length = fixed_length
        else:
            min_length = kwds.get('min_length', DEFAULT_MIN_SEQLEN)
            max_length = kwds.get('max_length', DEFAULT_MAX_SEQLEN)

        def generator() -> typing.Mapping[_T, _U]:
            length = random.randint(min_length, max_length)
            return {keygen(): valuegen() for _ in range(length)}

        return generator

    def listgen(value_type: typing.Type[_T], **kwds):
        """Generate sequence of values as list."""

        seq_generator = _seqgen(value_type, **kwds)

        def generator() -> typing.List[_T]:
            return list(seq_generator())

        return generator

    def setgen(value_type: typing.Type[_T], **kwds):
        """Generate sequence of values as set."""

        seq_generator = _seqgen(value_type, **kwds)

        def generator() -> typing.Set[_T]:
            return set(seq_generator())

        return generator

    def frozensetgen(value_type: typing.Type[_T], **kwds):
        """Generate sequence of values as frozenset."""

        seq_generator = _seqgen(value_type, **kwds)

        def generator() -> typing.FrozenSet[_T]:
            return frozenset(seq_generator())

        return generator

    def dequegen(value_type: typing.Type[_T], **kwds):
        """Generate sequence of values as deque."""

        seq_generator = _seqgen(value_type, **kwds)

        def generator() -> typing.Deque[_T]:
            return deque(seq_generator())

        return generator

    def dictgen(key_type: typing.Type[_T], value_type: typing.Type[_U], **kwds):
        """Generate mapping of keys and values as dict."""

        map_generator = _mapgen(key_type, value_type, **kwds)

        def generator() -> typing.Dict[_T, _U]:
            return dict(map_generator())

        return generator

    PROTO_GENERATORS = {
        bool: boolgen,
        int: intgen,
        float: floatgen,
        str: strgen,
        bytes: bytesgen,
        list: listgen,
        set: setgen,
        frozenset: frozensetgen,
        deque: dequegen,
        dict: dictgen,
    }

    GENERATORS = {
        # INT
        U8: intgen(0, (1 << 8) - 1),
        U16: intgen(0, (1 << 16) - 1),
        U32: intgen(0, (1 << 32) - 1),
        U64: intgen(0, (1 << 64) - 1),
        U128: intgen(0, (1 << 128) - 1),
        I8: intgen(-(1 << 7), (1 << 7) - 1),
        I16: intgen(-(1 << 15), (1 << 15) - 1),
        I32: intgen(-(1 << 31), (1 << 31) - 1),
        I64: intgen(-(1 << 63), (1 << 63) - 1),
        I128: intgen(-(1 << 127), (1 << 127) - 1),

        # FLOAT
        F32: structgen(32, 'f'),
        F64: structgen(64, 'd'),
    }

    INITIALIZED_LOGGER = False

    def log_error(error, testname, arguments, exc_info):
        """Log error, and if not present, seed."""

        global INITIALIZED_LOGGER
        testdir = '.test_harness'

        if not os.path.exists(testdir):
            os.mkdir(testdir)
        if not INITIALIZED_LOGGER:
            now = datetime.datetime.now()
            strtime = now.strftime("%Y-%m-%d-%H-%M-%S")
            logging.basicConfig(filename=f'{testdir}/random-{strtime}.log')
            INITIALIZED_LOGGER = True
            with open(f'{testdir}/random-{strtime}.seed', 'w') as f:
                f.write(str(random.getstate()))

        argstr = ', '.join(f'{k}={v!r}' for k, v in arguments.items())
        logging.error(
            f'Exception in randomly-generated unittest'
            f' {testname}, arguments were: ({argstr})',
            exc_info=exc_info
        )

    def randomize_function(f, gv, lv, **kwds):
        """Determine the type signature of the function and provide random data."""

        # Unwrap our annotations to determine our argument types.
        calls = kwds.get('calls', DEFAULT_CALLS)
        args = f.__annotations__.copy()
        args.pop('self', None)
        args.pop('return', None)
        args = {k: eval(v, gv, lv) if isinstance(v, str) else v for k, v in args.items()}
        generators = {n: _generator(t, kwds.get(n)) for n, t in args.items()}

        @functools.wraps(f)
        def test(self):
            for _ in range(calls):
                arguments = {k: v() for k, v in generators.items()}
                try:
                    f(self, **arguments)
                except Exception as error:
                    testname = f'{self.__class__.__name__}.{f.__name__}'
                    exc_info = sys.exc_info()
                    log_error(error, testname, arguments, exc_info)
                    raise

        return test

    def randomize_kwds(gv, lv, kwds):
        """Generator a function decorator with the bound keywords."""

        def decorator(f):
            return randomize_function(f, gv, lv, **kwds)

        return decorator

    def randomize(*args, **kwds):
        """Randomize arguments to a function call."""

        # Get global/local vars to evaluate forward refs.
        frame = inspect.stack()[1].frame
        gv = frame.f_globals
        lv = frame.f_locals

        # Process our arguments.
        if len(args) == 1 and callable(args[0]) and not kwds:
            return randomize_function(args[0], gv, lv)
        elif len(args):
            raise ValueError('randomize does not support positional arguments.')
        return randomize_kwds(gv, lv, kwds)



"""
    to_future
    =========

    RxPy extension to convert an observable to a future-like object.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import asyncio
import rx
from rx.internal import extensionmethod


def to_future(self) -> asyncio.Future:
    """Convert observable to an awaitable object."""

    future: asyncio.Future = asyncio.Future()
    values = []

    def on_next(v):
        values.append(v)

    def on_error(ex):
        future.set_exception(ex)

    def on_completed():
        # Fixes a critical bug where if no value is assigned, the
        # observable will hang forever.
        v = values.pop() if values else None
        future.set_result(v)

    self.subscribe(on_next, on_error, on_completed)

    return future


def __await__(self):
    return to_future(self).__await__()


extensionmethod(rx.AnonymousObservable)(__await__)
extensionmethod(rx.Observable)(__await__)



"""
    node_time
    =========

    Describes the time of the NEM node.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations

from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['NodeTime']


# TODO(ahuszagh) Add unittests.
@util.inherit_doc
@util.dataclass(frozen=True)
class NodeTime(util.DTO):
    """
    Node information.

    :param public_key: Public key of node.
    :param port: Port to communicate with node over.

    DTO Format:
        .. code-block:: yaml

            CommunicationTimestampsDTO:
                sendTimestamp: UInt64DTO
                receiveTimestamp: UInt64DTO

            NodeTimeDTO:
                communicationTimestamps: CommunicationTimestampsDTO
    """

    send_timestamp: int
    receive_timestamp: int

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'communicationTimestamps': {
                'sendTimestamp': util.u64_to_dto(self.send_timestamp),
                'receiveTimestamp': util.u64_to_dto(self.receive_timestamp),
            },
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        timestamps = data['communicationTimestamps']
        return cls(
            send_timestamp=util.u64_from_dto(timestamps['sendTimestamp']),
            receive_timestamp=util.u64_from_dto(timestamps['receiveTimestamp']),
        )



"""
    receipt
    ===========

    Abstract base class for transactions.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import bidict
import typing

from ..blockchain.network_type import OptionalNetworkType
from .format import DTOFormat
from .receipt_base import TypeMap, ReceiptBase
from ... import util

__all__ = ['Receipt']


@util.inherit_doc
class Receipt(ReceiptBase):
    """Abstract, non-embedded Receipt base class."""

    __slots__ = ()
    # Overridable classvars.
    TYPE_MAP: typing.ClassVar[TypeMap] = bidict.bidict()

    DTO: typing.ClassVar[DTOFormat] = DTOFormat(
        names={
            'type': 'type',
            'version': 'version',
        },
    )

    # DTO

    @classmethod
    def validate_dto_shared(cls, data: dict) -> bool:
        required_keys = {
            'version',
            'type',
        }
        return cls.validate_dto_required(data, required_keys)

    def to_dto_shared(
        self,
        network_type: OptionalNetworkType
    ) -> dict:
        # Shared data and callbacks.
        data: dict = {}
        cb = lambda k, v: self.DTO.save(k, data, v)
        cb_get = lambda k: cb(k, getattr(self, k))

        # Save shared data.
        # Do not export `network_type`, already exported
        # with version.
        cb_get('version')
        cb_get('type')

        return data

    def load_dto_shared(
        self,
        data: dict,
        network_type: OptionalNetworkType,
    ) -> None:
        # Shared data and callbacks.
        cb = lambda k: self.DTO.load(k, data)
        cb_set = lambda k: self._set(k, cb(k))

        # Load shared data.
        cb_set('version')
        cb_set('type')



"""
    dataclasses
    ===========

    Wrappers around Python's dataclasses to support slots and defaults.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import contextlib
import copy
import dataclasses
import inspect
import typing

__all__ = [
    'Field',
    'FrozenInstanceError',
    'InitVar',
    'MISSING',
    'dataclass',
    'is_dataclass',
]

# HELPERS

Field = dataclasses.Field
FrozenInstanceError = dataclasses.FrozenInstanceError
InitVar = dataclasses.InitVar
MISSING = dataclasses.MISSING
is_dataclass = dataclasses.is_dataclass
# ForwardRef isn't part of the public API.
# If it's not present, just skip it.
ForwardRef = getattr(typing, 'ForwardRef', str)
Vars = typing.Dict[str, typing.Any]
DictType = typing.Mapping[str, typing.Any]
TupleType = typing.Sequence[typing.Any]
DictFactory = typing.Callable[..., DictType]
TupleFactory = typing.Callable[..., TupleType]

# NEW METHODS


def get_argnames(func):
    """Get the argument names from a function."""

    # Get all positional arguments in __init__, including named
    # and named optional arguments. `co_varnames` stores all argument
    # names (including local variable names) in order, starting with
    # function arguments, so only grab `co_argcount` varnames.
    code = func.__code__
    argcount = code.co_argcount
    return code.co_varnames[:argcount]


def set_defaults(
    cls: typing.Type,
    defaults: Vars,
) -> None:
    """Set and validate optional default arguments."""

    if not defaults:
        return

    # Need to validate we aren't adding defaults for interior items.
    init = cls.__init__
    varnames = get_argnames(init)
    count = len(defaults)
    if not all(i in defaults for i in varnames[-count:]):
        raise SyntaxError("non-default argument follows default argument")
    if init.__defaults__ is not None:
        raise SyntaxError("__defaults__ should be none.")

    init.__defaults__ = tuple(defaults.values())


def fix_annotation(type, global_vars: Vars, local_vars: Vars):
    """Fix annotation for field with type."""

    if isinstance(type, str):
        try:
            # Silence a warning about sec, all the arguments passed.
            # Eval only gets passed for internal type annotations.
            return eval(type, global_vars, local_vars)  # nosec
        except NameError:
            return type
    elif isinstance(type, ForwardRef):
        arg = fix_annotation(type.__forward_arg__, global_vars, local_vars)
        is_argument = type.__forward_is_argument__
        if isinstance(arg, str):
            return ForwardRef(arg, is_argument)
        return arg
    elif hasattr(type, '__args__'):
        args = type.__args__
        args = tuple((fix_annotation(i, global_vars, local_vars) for i in args))
        type.__args__ = args

    return type


def fix_annotations(
    cls: typing.Type,
    clsdict: Vars,
    global_vars: Vars,
    local_vars: Vars,
) -> None:
    """Fix any forward references to variables defined in the callee scope."""

    # Don't care except when enforcing types.
    if typing.TYPE_CHECKING:
        annotations = clsdict['__annotations__']
        for field, type in annotations.items():
            type = fix_annotation(type, global_vars, local_vars)
            annotations[field] = type


def is_classvar(x, global_vars: Vars, local_vars: Vars) -> bool:
    """Determine if x is a ClassVar."""

    if isinstance(x, str):
        # Silence a warning about sec, all the arguments passed.
        # Eval only gets passed for internal type annotations.
        x = eval(x, global_vars, local_vars)    # nosec
    return getattr(x, '__origin__', None) is typing.ClassVar


def set_slots(
    cls: typing.Type,
    clsdict: Vars,
    slots: bool,
    global_vars: Vars,
    local_vars: Vars,
) -> None:
    """Set default __slots__ implementation."""

    if not slots or '__slots__' in clsdict:
        return
    annotations = clsdict['__annotations__']
    is_cv = lambda x: is_classvar(x, global_vars, local_vars)
    slots = (k for k, v in annotations.items() if not is_cv(v))
    clsdict['__slots__'] = tuple(slots)


def set_copy(
    cls: typing.Type,
    clsdict: Vars,
    copy: bool,
) -> None:
    """Set default __copy__ implementation."""

    if not copy or '__copy__' in clsdict:
        return

    def func(self):
        return type(self)(**replace_dict(self))

    func.__name__ = '__copy__'
    func.__qualname__ = f'{cls.__qualname__}.__copy__'
    func.__module__ = cls.__module__
    clsdict['__copy__'] = func


def set_deepcopy(
    cls: typing.Type,
    clsdict: Vars,
    deepcopy: bool,
) -> None:
    """Set default __deepcopy__ implementation."""

    if not deepcopy or '__deepcopy__' in clsdict:
        return

    def func(self, memo=None):
        data = copy.deepcopy(replace_dict(self), memo)
        return type(self)(**data)

    func.__name__ = '__deepcopy__'
    func.__qualname__ = f'{cls.__qualname__}.__deepcopy__'
    func.__module__ = cls.__module__
    clsdict['__deepcopy__'] = func


def shallow_asdict(self, dict_factory: DictFactory = dict) -> DictType:
    names = [i.name for i in dataclasses.fields(self)]
    return dict_factory([(i, getattr(self, i)) for i in names])


def deep_asdict(self, dict_factory: DictFactory = dict) -> DictType:
    return dataclasses.asdict(self, dict_factory=dict_factory)


def set_asdict(
    cls: typing.Type,
    clsdict: Vars,
    asdict: bool,
) -> None:
    """Set default asdict implementation."""

    if not asdict or 'asdict' in clsdict:
        return

    def func(
        self,
        recurse: bool = True,
        dict_factory: DictFactory = dict,
    ) -> DictType:
        if recurse:
            return deep_asdict(self, dict_factory=dict_factory)
        return shallow_asdict(self, dict_factory=dict_factory)

    func.__name__ = 'asdict'
    func.__qualname__ = f'{cls.__qualname__}.asdict'
    func.__module__ = cls.__module__
    clsdict['asdict'] = func


def shallow_astuple(self, tuple_factory: TupleFactory = tuple) -> TupleType:
    names = [i.name for i in dataclasses.fields(self)]
    return tuple_factory([getattr(self, i) for i in names])


def deep_astuple(self, tuple_factory: TupleFactory = tuple) -> TupleType:
    return dataclasses.astuple(self, tuple_factory=tuple_factory)


def set_astuple(
    cls: typing.Type,
    clsdict: Vars,
    astuple: bool,
) -> None:
    """Set default astuple implementation."""

    if not astuple or 'astuple' in clsdict:
        return

    def func(
        self,
        recurse: bool = True,
        tuple_factory: TupleFactory = tuple,
    ) -> TupleType:
        if recurse:
            return deep_astuple(self, tuple_factory=tuple_factory)
        return shallow_astuple(self, tuple_factory=tuple_factory)

    func.__name__ = 'astuple'
    func.__qualname__ = f'{cls.__qualname__}.astuple'
    func.__module__ = cls.__module__
    clsdict['astuple'] = func


def set_fields(
    cls: typing.Type,
    clsdict: Vars,
    fields: bool,
) -> None:
    """Set default fields implementation."""

    if not fields or 'fields' in clsdict:
        return

    def func(self) -> tuple:
        return dataclasses.fields(self)

    func.__name__ = 'fields'
    func.__qualname__ = f'{cls.__qualname__}.fields'
    func.__module__ = cls.__module__
    clsdict['fields'] = func


def replace_dict(self):
    # This is a smart-replace, any fields that are not defined
    # in the initializer are ignored, since it is assumed
    # sensible defaults are automatically provided for those.
    varnames = get_argnames(self.__init__.__func__)[1:]
    return {k: getattr(self, k) for k in varnames}


def set_replace(
    cls: typing.Type,
    clsdict: Vars,
    replace: bool,
) -> None:
    """Set default replace implementation."""

    if not replace or 'replace' in clsdict:
        return

    def func(self, **changes):
        # Convert to a dictionary and then call __init__.
        asdict = replace_dict(self)
        asdict.update(changes)
        return self.__class__(**asdict)

    func.__name__ = 'replace'
    func.__qualname__ = f'{cls.__qualname__}.replace'
    func.__module__ = cls.__module__
    clsdict['replace'] = func


def set_miscellaneous(cls: typing.Type, clsdict: Vars) -> None:
    """Set miscellaneous data for the class."""

    clsdict['_set'] = object.__setattr__


# DATACLASS METACLASS


def wrap_dataclass(
    cls: typing.Type,
    global_vars: typing.Dict[str, typing.Any],
    local_vars: typing.Dict[str, typing.Any],
    slots: bool = True,
    copy: bool = True,
    deepcopy: bool = True,
    asdict: bool = True,
    astuple: bool = True,
    fields: bool = True,
    replace: bool = True,
) -> typing.Type:
    """Wrap a dataclass base with the desired methods."""

    mcls = cls.__class__
    name = cls.__name__
    bases = cls.__bases__
    clsdict = cls.__dict__.copy()
    clsdict.setdefault('__annotations__', {})
    fix_annotations(cls, clsdict, global_vars, local_vars)
    set_slots(cls, clsdict, slots, global_vars, local_vars)
    set_copy(cls, clsdict, copy)
    set_deepcopy(cls, clsdict, deepcopy)
    set_asdict(cls, clsdict, asdict)
    set_astuple(cls, clsdict, astuple)
    set_fields(cls, clsdict, fields)
    set_replace(cls, clsdict, replace)
    set_miscellaneous(cls, clsdict)

    return mcls.__new__(mcls, name, bases, clsdict)


# PATCHES


def update_method(
    cls: typing.Type,
    new_cls: typing.Type,
    func,
) -> None:
    """
    Due to our dynamic creation of a new class, our old class may still
    be present in some function closures, through `super()/__class__`.
    We should not have remnants of the old class anywhere else, since
    that would require hard-coding the actual class name, which should
    not be done for obvious reasons.

    To rectify this, we can check if '__class__' is in nonlocal vars
    for each function (`func.__code__.co_freevars`), and if it is,
    and the old class is bound, update it.

    This is all well-documented in the Python data model:
        __code__: Contains compiled function bytecode.
        co_freevars: Contains free variables referenced inside closure.
        __closure__: None or tuple of cells for the functions free variables.
        cell_contents: Get value of cell.

    Since we want a value injected locally (a free variable), with the
    name `__class__`, we can use these attributes to determine if the old
    class is bound to `__class__`, and if so, overwrite it.
    """

    code = func.__code__
    closure = func.__closure__
    freevars = code.co_freevars
    if closure and freevars and '__class__' in freevars:
        # Have a specified class in freevars, must fix in closure.
        # Only change if the cell_contents is the old cls,
        # which we need to replace with the new cls.
        for cell in closure:
            if cell.cell_contents is cls:
                cell.cell_contents = new_cls


def update_methods(cls: typing.Type, new_cls: typing.Type) -> None:
    """Replace all instances of `super()/__class__` with the new class."""

    funcs = inspect.getmembers(new_cls, inspect.isroutine)
    for _, func in funcs:
        # Unwrap method if applicable.
        func = getattr(func, '__func__', func)
        with contextlib.suppress(AttributeError):
            # Builtin functions won't have __code__.
            update_method(cls, new_cls, func)


def update_classvars(cls: typing.Type, new_cls: typing.Type) -> None:
    """Replace all instances of the old class in class variables."""

    # We're going to cheat, since this is a painfully long process.
    # We know the only classvars that can be present are in:
    #   1. TYPE_MAP

    # Mapping of transaction type to transactions.
    # We use a bidict to get O(1) inverse lookup times.
    type_map = getattr(cls, 'TYPE_MAP', None)
    if type_map is not None and cls in type_map.inverse:
        type = type_map.inverse.pop(cls)
        type_map[type] = new_cls


def update_closure(cls: typing.Type, new_cls: typing.Type) -> None:
    """
    Due to our dynamic creation of a new class, our old class
    may be present in our new class definition.

    Most notably, it may be present in:
        1. Methods using `super()/__class__`.
        2. Any class variables.
    """

    update_methods(cls, new_cls)
    update_classvars(cls, new_cls)


# DATACLASS


def dataclass(
    cls: typing.Optional[typing.Type] = None,
    *,
    init: bool = True,
    repr: bool = True,
    eq: bool = True,
    order: bool = False,
    unsafe_hash: bool = False,
    frozen: bool = False,
    slots: bool = True,
    copy: bool = True,
    deepcopy: bool = True,
    asdict: bool = True,
    astuple: bool = True,
    fields: bool = True,
    replace: bool = True,
    **defaults: typing.Any
):
    """Generate a slotted dataclass with optional default arguments."""

    frame = inspect.stack()[1].frame
    global_vars = frame.f_globals
    local_vars = frame.f_locals
    dataclass_kwds = {
        'init': init,
        'repr': repr,
        'eq': eq,
        'order': order,
        'unsafe_hash': unsafe_hash,
        'frozen': frozen,
    }
    wrap_kwds = {
        'slots': slots,
        'copy': copy,
        'deepcopy': deepcopy,
        'asdict': asdict,
        'astuple': astuple,
        'fields': fields,
        'replace': replace,
    }

    def wrap(cls: typing.Type) -> typing.Type:
        base = wrap_dataclass(cls, global_vars, local_vars, **wrap_kwds)
        new_cls = dataclasses.dataclass(**dataclass_kwds)(base)
        set_defaults(new_cls, defaults)
        update_closure(cls, new_cls)
        return new_cls

    if cls is None:
        return wrap
    return wrap(cls)



"""
    transaction
    ===========

    Abstract base class for transactions.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import bidict
import typing

from .base import TransactionBase, TypeMap
from .format import CatbufferFormat, DTOFormat
from .inner_transaction import InnerTransaction
from .signed_transaction import SignedTransaction
from ..account.account import Account
from ..account.public_account import PublicAccount
from ..blockchain.network_type import NetworkType
from ... import util

__all__ = ['Transaction']


@util.inherit_doc
class Transaction(TransactionBase):
    """Abstract, non-embedded transaction base class."""

    __slots__ = ()
    # Overridable classvars.
    TYPE_MAP: typing.ClassVar[TypeMap] = bidict.bidict()
    CATBUFFER: typing.ClassVar[CatbufferFormat] = CatbufferFormat(
        # Layout
        #   uint32_t size
        #   uint8_t[64] signature
        #   uint8_t[32] signer
        #   uint8_t[3] version
        #   uint8_t network_type
        #   uint16_t type
        #   uint64_t max_fee
        #   uint64_t deadline
        slices={
            # Main fields.
            'size': slice(0, 4),
            'signature': slice(4, 68),
            'signer': slice(68, 100),
            'version': slice(100, 103),
            'network_type': slice(103, 104),
            'type': slice(104, 106),
            'max_fee': slice(106, 114),
            'deadline': slice(114, 122),
            # Helpers.
            'signature_half': slice(4, 36),
            'signing_bytes': slice(100, None),
        },
        size_shared=122,
    )
    DTO: typing.ClassVar[DTOFormat] = DTOFormat(
        names={
            'signature': 'signature',
            'signer': 'signer',
            'version': 'version',
            'network_type': 'version',
            'type': 'type',
            'max_fee': 'maxFee',
            'deadline': 'deadline',
            'transaction_info': 'meta',
        },
    )

    # SIGNING

    def sign_with(
        self,
        account: Account,
        gen_hash: typing.AnyStr,
        fee_strategy: util.FeeCalculationStrategy = util.FeeCalculationStrategy.MEDIUM,
    ) -> SignedTransaction:
        """
        Serialize and sign transaction.

        :param account: Account to sign transaction.
        :param gen_hash: Nemesis generation hash.
        :return: Signed transaction data,
        """

        # Serialize transaction data, sign, and generate a hash.
        transaction = self.to_catbuffer(fee_strategy=fee_strategy)
        payload = account.sign(transaction, gen_hash)  # type: ignore

        hash = self.transaction_hash(payload, gen_hash)  # type: ignore
        return SignedTransaction(  # type: ignore
            payload,
            hash,
            account.public_key,
            self.type,
            self.network_type
        )

    @staticmethod
    def transaction_hash(transaction: typing.AnyStr, gen_hash: typing.AnyStr) -> str:
        """Generate transaction hash from signed transaction payload."""

        # Decode the data.
        transaction = util.decode_hex(transaction, with_prefix=True)

        # Extract the first half of the signature, the signer, and the
        # remaining signing bytes, sign them, and get the hash.
        slices = Transaction.CATBUFFER.slices

        signature_half = transaction[slices['signature_half']]
        signer = transaction[slices['signer']]
        rest = transaction[slices['signing_bytes']]

        data = signature_half + signer + util.decode_hex(gen_hash, with_prefix=True) + rest
        hash = util.hashlib.sha3_256(data).hexdigest()

        return typing.cast(str, hash)

    # AGGREGATE

    def aggregate_transaction(
        self,
        signer: typing.Optional[PublicAccount] = None,
    ) -> bytes:
        """Build aggregate transaction."""

        if signer is None:
            signer = self.signer
        if signer is None:
            raise ValueError('Cannot create aggregate transaction without signer.')

        transaction = self.to_aggregate(signer).to_catbuffer()
        return typing.cast(bytes, transaction)

    def to_aggregate(self, signer: PublicAccount) -> InnerTransaction:
        """Convert transaction to inner transaction."""
        return InnerTransaction.create_from_transaction(self, signer)

    # CATBUFFER

    def catbuffer_size_specific(self) -> int:
        raise util.AbstractMethodError

    def to_catbuffer_shared(
        self,
        network_type: NetworkType,
    ) -> bytes:
        # Shared data and callbacks.
        data = bytearray(self.catbuffer_size_shared())
        cb = lambda k, v: self.CATBUFFER.save(k, data, v, network_type)
        cb_get = lambda k: cb(k, getattr(self, k))

        # Save shared data.
        cb('size', self.catbuffer_size())
        cb_get('signature')
        cb_get('signer')
        cb_get('version')
        cb_get('network_type')
        cb_get('type')
        cb_get('max_fee')
        cb_get('deadline')

        return bytes(data)

    def load_catbuffer_shared(
        self,
        data: bytes,
        network_type: NetworkType,
    ) -> bytes:
        # Shared data and callbacks.
        cb = lambda k: self.CATBUFFER.load(k, data, network_type)
        cb_set = lambda k: self._set(k, cb(k))

        # Load shared data.
        cb_set('signature')
        cb_set('signer')
        cb_set('version')
        self._set('network_type', network_type)
        cb_set('type')
        cb_set('max_fee')
        cb_set('deadline')
        self._set('transaction_info', None)

        return data[self.catbuffer_size_shared():]

    # DTO

    @classmethod
    def validate_dto_shared(cls, data: dict) -> bool:
        required_keys = {
            'version',
            'type',
        }
        return cls.validate_dto_required(data, required_keys)

    def to_dto_shared(
        self,
        network_type: NetworkType,
    ) -> dict:
        # Shared data and callbacks.
        data: dict = {}
        cb = lambda k, v: self.DTO.save(k, data, v, network_type)
        cb_get = lambda k: cb(k, getattr(self, k))

        # Save shared data.
        # Do not export `network_type`, already exported
        # with version.
        cb_get('signature')
        cb_get('signer')
        cb_get('version')
        cb_get('type')
        cb_get('max_fee')
        cb_get('deadline')
        cb_get('transaction_info')

        return data

    def load_dto_shared(
        self,
        data: dict,
        network_type: NetworkType,
    ) -> None:
        # Shared data and callbacks.
        cb = lambda k: self.DTO.load(k, data, network_type)
        cb_set = lambda k: self._set(k, cb(k))

        # Load shared data.
        cb_set('signature')
        cb_set('signer')
        cb_set('version')
        self._set('network_type', network_type)
        cb_set('type')
        cb_set('max_fee')
        cb_set('deadline')
        cb_set('transaction_info')




from __future__ import annotations

from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['CatapultUpgrade', 'Upgrade']


@util.inherit_doc
@util.dataclass(frozen=True)
class Upgrade(util.DTO):

    height: int
    blockchain_version: int

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_l1 = {'height', 'blockChainVersion'}
        return (
            # Level 1
            cls.validate_dto_required(data, required_l1)
            and cls.validate_dto_all(data, required_l1)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'height': util.u64_to_dto(self.height),
            'blockChainVersion': util.u64_to_dto(self.blockchain_version),
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        return cls(
            height=util.u64_from_dto(data['height']),
            blockchain_version=util.u64_from_dto(data['blockChainVersion']),
        )


@util.inherit_doc
@util.dataclass(frozen=True)
class CatapultUpgrade(util.DTO):
    """
    Basic information describing a blockchain upgrade.

    :param blockchain_upgrade: Blockchain upgrade.

    """

    blockchain_upgrade: Upgrade

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_l1 = {'blockchainUpgrade'}
        return (
            # Level 1
            cls.validate_dto_required(data, required_l1)
            and cls.validate_dto_all(data, required_l1)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'blockchainUpgrade': self.blockchain_upgrade.to_dto(),
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        return cls(
            blockchain_upgrade=Upgrade.create_from_dto(data['blockchainUpgrade']),
        )



# Test our randomization test suite to ensure it is properly
# generating random data for our arguments.

from collections import deque
import enum
import math
import re
import string
import typing

from tests import harness


class Colors(enum.IntEnum):
    RED = 1 << 16
    GREEN = 1 << 8
    BLUE = 1


class TestRandomize(harness.TestCase):

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)

    @harness.randomize
    def test_noargs_nokwds(self):
        pass

    @harness.randomize
    def test_enum(self, x: Colors):
        self.assertIsInstance(x, Colors)

    @harness.randomize
    def test_bool_nokwds(self, x: bool):
        self.assertIsInstance(x, bool)

    @harness.randomize
    def test_int_nokwds(self, x: int):
        self.assertIsInstance(x, int)

    @harness.randomize(x={'min_value': 5, 'max_value': 10})
    def test_int_range(self, x: int):
        self.assertIsInstance(x, int)
        self.assertGreaterEqual(x, 5)
        self.assertLessEqual(x, 10)

    @harness.randomize
    def test_u8(self, x: harness.U8):
        self.assertIsInstance(x, int)
        self.assertGreaterEqual(x, 0)
        self.assertLess(x, 1 << 8)

    @harness.randomize
    def test_u16(self, x: harness.U16):
        self.assertIsInstance(x, int)
        self.assertGreaterEqual(x, 0)
        self.assertLess(x, 1 << 16)

    @harness.randomize
    def test_u32(self, x: harness.U32):
        self.assertIsInstance(x, int)
        self.assertGreaterEqual(x, 0)
        self.assertLess(x, 1 << 32)

    @harness.randomize
    def test_u64(self, x: harness.U64):
        self.assertIsInstance(x, int)
        self.assertGreaterEqual(x, 0)
        self.assertLess(x, 1 << 64)

    @harness.randomize
    def test_u128(self, x: harness.U128):
        self.assertIsInstance(x, int)
        self.assertGreaterEqual(x, 0)
        self.assertLess(x, 1 << 128)

    @harness.randomize
    def test_i8(self, x: harness.I8):
        self.assertIsInstance(x, int)
        self.assertGreaterEqual(x, -(1 << 7))
        self.assertLess(x, 1 << 7)

    @harness.randomize
    def test_i16(self, x: harness.I16):
        self.assertIsInstance(x, int)
        self.assertGreaterEqual(x, -(1 << 15))
        self.assertLess(x, 1 << 15)

    @harness.randomize
    def test_i32(self, x: harness.I32):
        self.assertIsInstance(x, int)
        self.assertGreaterEqual(x, -(1 << 31))
        self.assertLess(x, 1 << 31)

    @harness.randomize
    def test_i64(self, x: harness.I64):
        self.assertIsInstance(x, int)
        self.assertGreaterEqual(x, -(1 << 63))
        self.assertLess(x, 1 << 63)

    @harness.randomize
    def test_i128(self, x: harness.I128):
        self.assertIsInstance(x, int)
        self.assertGreaterEqual(x, -(1 << 127))
        self.assertLess(x, 1 << 127)

    @harness.randomize
    def test_float_nokwds(self, x: float):
        self.assertIsInstance(x, float)

    @harness.randomize
    def test_f32(self, x: harness.F32):
        self.assertIsInstance(x, float)
        self.assertTrue(
            not math.isfinite(x)
            or x == 0.0
            or 1e-45 <= abs(x) <= 3.4028235e+38
        )

    @harness.randomize
    def test_f64(self, x: harness.F64):
        self.assertIsInstance(x, float)
        self.assertTrue(
            not math.isfinite(x)
            or x == 0.0
            or 5e-324 <= abs(x) <= 1.7976931348623157e+308
        )

    @harness.randomize
    def test_str_nokwds(self, x: str):
        self.assertIsInstance(x, str)

    @harness.randomize(x={'pattern': r'\w{5,10}'})
    def test_str_pattern(self, x: str):
        self.assertIsInstance(x, str)
        self.assertGreaterEqual(len(x), 5)
        self.assertLessEqual(len(x), 10)
        self.assertTrue(re.match(r'\w{5,10}', x))

    @harness.randomize(x={'letters': string.hexdigits})
    def test_str_letters(self, x: str):
        self.assertIsInstance(x, str)
        self.assertTrue(re.match(r'[A-Fa-f0-9]+', x))

    @harness.randomize(x={'min_length': 3, 'max_length': 4})
    def test_str_min_max_length(self, x: str):
        self.assertIsInstance(x, str)
        self.assertTrue(len(x) in (3, 4))

    @harness.randomize(x={'fixed_length': 10})
    def test_str_fixed_length(self, x: str):
        self.assertIsInstance(x, str)
        self.assertEqual(len(x), 10)

    @harness.randomize
    def test_bytes_nokwds(self, x: bytes):
        self.assertIsInstance(x, bytes)

    @harness.randomize(x={'letters': string.hexdigits.encode('ascii')})
    def test_bytes_letters(self, x: bytes):
        self.assertIsInstance(x, bytes)
        self.assertTrue(re.match(rb'[A-Fa-f0-9]+', x))

    @harness.randomize(x={'min_length': 3, 'max_length': 4})
    def test_bytes_min_max_length(self, x: bytes):
        self.assertIsInstance(x, bytes)
        self.assertTrue(len(x) in (3, 4))

    @harness.randomize(x={'fixed_length': 10})
    def test_bytes_fixed_length(self, x: bytes):
        self.assertIsInstance(x, bytes)
        self.assertEqual(len(x), 10)

    @harness.randomize
    def test_list_nokwds(self, x: typing.List[int]):
        self.assertIsInstance(x, list)
        for xi in x:
            self.assertIsInstance(xi, int)

    @harness.randomize(x={'value_type': int})
    def test_list_value_type(self, x: list):
        self.assertIsInstance(x, list)
        for xi in x:
            self.assertIsInstance(xi, int)

    @harness.randomize(x={'value': {'min_value': 5, 'max_value': 10}})
    def test_list_args(self, x: typing.List[int]):
        self.assertIsInstance(x, list)
        for xi in x:
            self.assertIsInstance(xi, int)
            self.assertGreaterEqual(xi, 5)
            self.assertLessEqual(xi, 10)

    @harness.randomize
    def test_set_nokwds(self, x: typing.Set[int]):
        self.assertIsInstance(x, set)
        for xi in x:
            self.assertIsInstance(xi, int)

    @harness.randomize
    def test_frozenset_kwds(self, x: typing.FrozenSet[int]):
        self.assertIsInstance(x, frozenset)
        for xi in x:
            self.assertIsInstance(xi, int)

    @harness.randomize
    def test_deque_nokwds(self, x: typing.Deque[int]):
        self.assertIsInstance(x, deque)
        for xi in x:
            self.assertIsInstance(xi, int)

    @harness.randomize
    def test_dict_nokwds(self, x: typing.Dict[str, int]):
        self.assertIsInstance(x, dict)
        for ki, vi in x.items():
            self.assertIsInstance(ki, str)
            self.assertIsInstance(vi, int)

    @harness.randomize(x={
        'key': {'fixed_length': 0},
        'value': {'min_value': 0, 'max_value': 0},
    })
    def test_dict_kwds(self, x: typing.Dict[str, int]):
        self.assertIsInstance(x, dict)
        for ki, vi in x.items():
            self.assertEqual(ki, '')
            self.assertEqual(vi, 0)


class TestRandomizeForwardRef(harness.TestCase):

    def __init__(self, *args, **kwds):
        super().__init__(*args, **kwds)

    @harness.randomize
    def test_noargs_nokwds(self):
        pass

    @harness.randomize
    def test_enum(self, x: 'Colors'):
        self.assertIsInstance(x, Colors)

    @harness.randomize
    def test_bool_nokwds(self, x: 'bool'):
        self.assertIsInstance(x, bool)

    @harness.randomize
    def test_int_nokwds(self, x: 'int'):
        self.assertIsInstance(x, int)

    @harness.randomize
    def test_float_nokwds(self, x: 'float'):
        self.assertIsInstance(x, float)

    @harness.randomize
    def test_str_nokwds(self, x: 'str'):
        self.assertIsInstance(x, str)

    @harness.randomize
    def test_bytes_nokwds(self, x: 'bytes'):
        self.assertIsInstance(x, bytes)

    @harness.randomize
    def test_list_nokwds(self, x: 'typing.List[int]'):
        self.assertIsInstance(x, list)
        for xi in x:
            self.assertIsInstance(xi, int)


class TestInvalidRandomize(harness.TestCase):

    def test_invalid_list(self):
        with self.assertRaises(TypeError):
            @harness.randomize
            def test(self, x: list):
                pass

    def test_invalid_set(self):
        with self.assertRaises(TypeError):
            @harness.randomize
            def test(self, x: set):
                pass

    def test_invalid_frozenset(self):
        with self.assertRaises(TypeError):
            @harness.randomize
            def test(self, x: frozenset):
                pass

    def test_invalid_deque(self):
        with self.assertRaises(TypeError):
            @harness.randomize
            def test(self, x: deque):
                pass

    def test_invalid_dict(self):
        with self.assertRaises(TypeError):
            @harness.randomize
            def test(self, x: dict):
                pass


class TestInvalidRandomizeFutureRef(harness.TestCase):

    def test_invalid_list(self):
        with self.assertRaises(TypeError):
            @harness.randomize
            def test(self, x: 'list'):
                pass

    def test_invalid_set(self):
        with self.assertRaises(TypeError):
            @harness.randomize
            def test(self, x: 'set'):
                pass

    def test_invalid_frozenset(self):
        with self.assertRaises(TypeError):
            @harness.randomize
            def test(self, x: 'frozenset'):
                pass

    def test_invalid_deque(self):
        with self.assertRaises(TypeError):
            @harness.randomize
            def test(self, x: 'deque'):
                pass

    def test_invalid_dict(self):
        with self.assertRaises(TypeError):
            @harness.randomize
            def test(self, x: 'dict'):
                pass




from __future__ import annotations

import typing

from ..account.address import Address
from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['ContractInfo']


@util.inherit_doc
@util.dataclass(frozen=True)
class ContractInfo(util.DTO):
    """

    :param contract: Contract info.

    DTO Format:
        .. code-block:: yaml

            ContractDTO:
                multisig: string
                multisigAddress: string
                start: UInt64DTO
                duration: UInt64DTO
                hash: string
                customers: string[]
                executors: string[]
                verifiers: string[]

            ContractInfoDTO:
                contract: ConfigDTO
    """

    multisig: str
    multisig_address: Address
    start: int
    duration: int
    hash: str
    customers: typing.Sequence[str]
    executors: typing.Sequence[str]
    verifiers: typing.Sequence[str]

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_l1 = {'contract'}
        required_l2 = {
            'multisig',
            'multisigAddress',
            'start',
            'duration',
            'hash',
            'customers',
            'executors',
            'verifiers'
        }
        return (
            # Level 1
            cls.validate_dto_required(data, required_l1)
            and cls.validate_dto_all(data, required_l1)
            # Level 2
            and cls.validate_dto_required(data['contract'], required_l2)
            and cls.validate_dto_all(data['contract'], required_l2)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        contract = {
            'multisig': self.multisig,
            'multisigAddress': util.hexlify(self.multisig_address.encoded),
            'start': util.u64_to_dto(self.start),
            'duration': util.u64_to_dto(self.duration),
            'hash': self.hash,
            'customers': self.customers,
            'executors': self.executors,
            'verifiers': self.verifiers
        }

        return {
            'contract': contract,
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        contract = data['contract']
        return cls(
            multisig=contract['multisig'],
            multisig_address=Address.create_from_encoded(contract['multisigAddress']),
            start=util.u64_from_dto(contract.get('start', [0, 0])),
            duration=util.u64_from_dto(contract.get('duration', [0, 0])),
            hash=contract['hash'],
            customers=contract['customers'],
            executors=contract['executors'],
            verifiers=contract['verifiers']
        )



"""
    aggregate_transaction_cosignature
    =================================

    Aggregate transaction signer and signature.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from ..account.public_account import PublicAccount
from ..blockchain.network_type import OptionalNetworkType
from ... import util

__all__ = ['AggregateTransactionCosignature']

SIZE = PublicAccount.CATBUFFER_SIZE + 64 * util.U8_BYTES


@util.inherit_doc
@util.dataclass(frozen=True)
class AggregateTransactionCosignature(util.Model):
    """
    Aggregate transaction signer and signature.

    :param signature: Signature of aggregate transaction done by cosigner.
    :param signer: Cosigner account.
    """

    signature: str
    signer: PublicAccount
    CATBUFFER_SIZE: typing.ClassVar[int] = SIZE

    def __init__(
        self,
        signature: typing.AnyStr,
        signer: PublicAccount,
    ) -> None:
        signature = util.encode_hex(signature)
        if len(signature) != 128:
            raise ValueError("Invalid signature length")
        self._set('signature', signature)
        self._set('signer', signer)

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_keys = {'signature', 'signer'}
        return (
            cls.validate_dto_required(data, required_keys)
            and cls.validate_dto_all(data, required_keys)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'signature': self.signature,
            'signer': self.signer.public_key,
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        signature = data['signature']
        signer = PublicAccount.create_from_public_key(data['signer'], network_type)
        return cls(signature, signer)

    def to_catbuffer(
        self,
        network_type: OptionalNetworkType = None,
        fee_strategy: util.FeeCalculationStrategy = util.FeeCalculationStrategy.MEDIUM,
    ) -> bytes:
        # uint8_t[32] signer
        # uint8_t[64] signature
        signer = util.unhexlify(self.signer.public_key)
        signature = util.unhexlify(self.signature)
        return signer + signature

    @classmethod
    def create_from_catbuffer(
        cls,
        data: bytes,
        network_type: OptionalNetworkType = None,
    ):
        # uint8_t[32] signer
        # uint8_t[64] signature
        signer = PublicAccount.create_from_public_key(data[:32], network_type)
        signature = data[32:96]
        return cls(signature, signer)  # type: ignore



"""
    mixin
    =====

    Mixins classes.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from . import abc
from . import documentation
from . import stdint

__all__ = [
    'EnumMixin',
    'IntMixin',
    'U8Mixin',
    'U16Mixin',
    'U32Mixin',
    'U64Mixin',
    'U128Mixin',
]

IntMixinType = typing.TypeVar('IntMixinType', bound='IntMixin')
U8MixinType = typing.TypeVar('U8MixinType', bound='U8Mixin')
U16MixinType = typing.TypeVar('U16MixinType', bound='U16Mixin')
U32MixinType = typing.TypeVar('U32MixinType', bound='U32Mixin')
U64MixinType = typing.TypeVar('U64MixinType', bound='U64Mixin')
U128MixinType = typing.TypeVar('U128MixinType', bound='U128Mixin')


class EnumMixin:
    """Mixin defining shared methods for enumerations."""

    __slots__ = ()

    def description(self) -> str:
        """Describe enumerated values in detail."""
        raise abc.AbstractMethodError


class IntMixin:
    """Mixin for classes that can be interpreted as integers."""

    __slots__ = ()

    def __int__(self) -> int:
        raise abc.AbstractMethodError

    def __index__(self) -> int:
        return self.__int__()

    def __format__(self, format_spec: str) -> str:
        return int(self).__format__(format_spec)

    @classmethod
    def create_from_hex(
        cls: typing.Type[IntMixinType],
        data: str
    ) -> IntMixinType:
        """
        Create instance of class from hex string.

        :param data: Hex-encoded ID data (with or without '0x' prefix).
        """

        # Assumes the number is encoded as if a 0x{hex} literal.
        # Same as the NEM2 library.
        return cls(int(data, 16))   # type: ignore


@documentation.inherit_doc
class U8Mixin(abc.Model):
    """Mixin for classes wrapping 8-bit integer types."""

    __slots__ = ()
    CATBUFFER_SIZE: typing.ClassVar[int] = stdint.U8_BYTES

    @classmethod
    def validate_dto(
        cls: typing.Type[U8MixinType],
        data: stdint.U8DTOType
    ) -> bool:
        """Validate the data-transfer object."""
        return isinstance(data, int) and 0 <= data < (1 << 8)

    def to_dto(
        self: U8MixinType,
        network_type: abc.OptionalNetworkType = None,
    ) -> stdint.U8DTOType:
        asint: int = int(self)      # type: ignore
        return stdint.u8_to_dto(asint)

    @classmethod
    def create_from_dto(
        cls: typing.Type[U8MixinType],
        data: stdint.U8DTOType,
        network_type: abc.OptionalNetworkType = None,
    ) -> U8MixinType:
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        asint = stdint.u8_from_dto(data)
        return cls(asint)           # type: ignore

    def to_catbuffer(
        self: U8MixinType,
        network_type: abc.OptionalNetworkType = None,
    ) -> bytes:
        asint: int = int(self)      # type: ignore
        return stdint.u8_to_catbuffer(asint)

    @classmethod
    def create_from_catbuffer(
        cls: typing.Type[U8MixinType],
        data: bytes,
        network_type: abc.OptionalNetworkType = None,
    ) -> U8MixinType:
        size = cls.CATBUFFER_SIZE
        asint = stdint.u8_from_catbuffer(data[:size])
        return cls(asint)           # type: ignore


@documentation.inherit_doc
class U16Mixin(abc.Model):
    """Mixin for classes wrapping 16-bit integer types."""

    __slots__ = ()
    CATBUFFER_SIZE: typing.ClassVar[int] = stdint.U16_BYTES

    @classmethod
    def validate_dto(
        cls: typing.Type[U16MixinType],
        data: stdint.U16DTOType
    ) -> bool:
        """Validate the data-transfer object."""
        return isinstance(data, int) and 0 <= data < (1 << 16)

    def to_dto(
        self: U16MixinType,
        network_type: abc.OptionalNetworkType = None,
    ) -> stdint.U16DTOType:
        asint: int = int(self)      # type: ignore
        return stdint.u16_to_dto(asint)

    @classmethod
    def create_from_dto(
        cls: typing.Type[U16MixinType],
        data: stdint.U16DTOType,
        network_type: abc.OptionalNetworkType = None,
    ) -> U16MixinType:
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        asint = stdint.u16_from_dto(data)
        return cls(asint)           # type: ignore

    def to_catbuffer(
        self: U16MixinType,
        network_type: abc.OptionalNetworkType = None,
    ) -> bytes:
        asint: int = int(self)      # type: ignore
        return stdint.u16_to_catbuffer(asint)

    @classmethod
    def create_from_catbuffer(
        cls: typing.Type[U16MixinType],
        data: bytes,
        network_type: abc.OptionalNetworkType = None,
    ) -> U16MixinType:
        size = cls.CATBUFFER_SIZE
        asint = stdint.u16_from_catbuffer(data[:size])
        return cls(asint)           # type: ignore


@documentation.inherit_doc
class U32Mixin(abc.Model):
    """Mixin for classes wrapping 32-bit integer types."""

    __slots__ = ()
    CATBUFFER_SIZE: typing.ClassVar[int] = stdint.U32_BYTES

    @classmethod
    def validate_dto(
        cls: typing.Type[U32MixinType],
        data: stdint.U32DTOType
    ) -> bool:
        """Validate the data-transfer object."""
        return isinstance(data, int) and 0 <= data < (1 << 32)

    def to_dto(
        self,
        network_type: abc.OptionalNetworkType = None,
    ) -> stdint.U32DTOType:
        asint: int = int(self)      # type: ignore
        return stdint.u32_to_dto(asint)

    @classmethod
    def create_from_dto(
        cls: typing.Type[U32MixinType],
        data: stdint.U32DTOType,
        network_type: abc.OptionalNetworkType = None,
    ) -> U32MixinType:
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        asint = stdint.u32_from_dto(data)
        return cls(asint)           # type: ignore

    def to_catbuffer(
        self: U32MixinType,
        network_type: abc.OptionalNetworkType = None,
    ) -> bytes:
        asint: int = int(self)      # type: ignore
        return stdint.u32_to_catbuffer(asint)

    @classmethod
    def create_from_catbuffer(
        cls: typing.Type[U32MixinType],
        data: bytes,
        network_type: abc.OptionalNetworkType = None,
    ) -> U32MixinType:
        size = cls.CATBUFFER_SIZE
        asint = stdint.u32_from_catbuffer(data[:size])
        return cls(asint)           # type: ignore


@documentation.inherit_doc
class U64Mixin(abc.Model):
    """Mixin for classes wrapping 64-bit integer types."""

    __slots__ = ()
    CATBUFFER_SIZE: typing.ClassVar[int] = stdint.U64_BYTES

    @classmethod
    def validate_dto(
        cls: typing.Type[U64MixinType],
        data: stdint.U64DTOType
    ) -> bool:
        """Validate the data-transfer object."""
        return len(data) == 2 and all(U32Mixin.validate_dto(i) for i in data)

    def to_dto(
        self: U64Mixin,
        network_type: abc.OptionalNetworkType = None,
    ) -> stdint.U64DTOType:
        asint: int = int(self)      # type: ignore
        return stdint.u64_to_dto(asint)

    @classmethod
    def create_from_dto(
        cls: typing.Type[U64MixinType],
        data: stdint.U64DTOType,
        network_type: abc.OptionalNetworkType = None,
    ) -> U64MixinType:
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        asint = stdint.u64_from_dto(data)
        return cls(asint)           # type: ignore

    def to_catbuffer(
        self: U64Mixin,
        network_type: abc.OptionalNetworkType = None,
    ) -> bytes:
        asint: int = int(self)      # type: ignore
        return stdint.u64_to_catbuffer(asint)

    @classmethod
    def create_from_catbuffer(
        cls: typing.Type[U64MixinType],
        data: bytes,
        network_type: abc.OptionalNetworkType = None,
    ) -> U64MixinType:
        size = cls.CATBUFFER_SIZE
        asint = stdint.u64_from_catbuffer(data[:size])
        return cls(asint)           # type: ignore


@documentation.inherit_doc
class U128Mixin(abc.Model):
    """Mixin for classes wrapping 128-bit integer types."""

    __slots__ = ()
    CATBUFFER_SIZE: typing.ClassVar[int] = stdint.U128_BYTES

    @classmethod
    def validate_dto(
        cls: typing.Type[U128MixinType],
        data: stdint.U128DTOType
    ) -> bool:
        """Validate the data-transfer object."""
        return len(data) == 2 and all(U64Mixin.validate_dto(i) for i in data)

    def to_dto(
        self: U128MixinType,
        network_type: abc.OptionalNetworkType = None,
    ) -> stdint.U128DTOType:
        asint: int = int(self)      # type: ignore
        return stdint.u128_to_dto(asint)

    @classmethod
    def create_from_dto(
        cls: typing.Type[U128MixinType],
        data: stdint.U128DTOType,
        network_type: abc.OptionalNetworkType = None,
    ) -> U128MixinType:
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        asint = stdint.u128_from_dto(data)
        return cls(asint)           # type: ignore

    def to_catbuffer(
        self: U128MixinType,
        network_type: abc.OptionalNetworkType = None,
    ) -> bytes:
        asint: int = int(self)      # type: ignore
        return stdint.u128_to_catbuffer(asint)

    @classmethod
    def create_from_catbuffer(
        cls: typing.Type[U128MixinType],
        data: bytes,
        network_type: abc.OptionalNetworkType = None,
    ) -> U128MixinType:
        size = cls.CATBUFFER_SIZE
        asint = stdint.u128_from_catbuffer(data[:size])
        return cls(asint)           # type: ignore



from xpxchain import models
from tests import harness


@harness.model_test_case({
    'type': models.AddressAlias,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'value': models.Address('SD5DT3CH4BLABL5HIMEKP2TAPUKF4NY3L5HRIR54'),
    },
    # TODO(ahuszagh) Check the format, and confirm.
    # https://xpx.slack.com/archives/CEZKUE4KB/p1553274126174200
    'dto': {
        'type': 2,
        'address': '90fa39ec47e05600afa74308a7ea607d145e371b5f4f1447bc',
    },
    'extras': {
        'mosaic_id': models.MosaicId(5),
    },
})
class TestAddressAlias(harness.TestCase):

    def test_rich_eq(self):
        empty_alias = models.Alias()
        address_alias = models.Alias(models.AliasType.ADDRESS, self.data['value'])
        mosaic_alias = models.Alias(models.AliasType.MOSAIC_ID, self.extras['mosaic_id'])

        self.assertNotEqual(self.model, empty_alias)
        self.assertEqual(self.model, address_alias)
        self.assertNotEqual(self.model, mosaic_alias)

    def test_properties(self):
        with self.assertRaises(ValueError):
            self.model.mosaic_id

    def test_from_invalid_dto(self):
        with self.assertRaises(ValueError):
            self.type.create_from_dto({'type': 0})


@harness.model_test_case({
    'type': models.Alias,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'type': models.AliasType.ADDRESS,
        'value': models.Address('SD5DT3CH4BLABL5HIMEKP2TAPUKF4NY3L5HRIR54'),
    },
    # TODO(ahuszagh) Check the format, and confirm.
    # https://xpx.slack.com/archives/CEZKUE4KB/p1553274126174200
    'dto': {
        'type': 2,
        'address': '90fa39ec47e05600afa74308a7ea607d145e371b5f4f1447bc',
    },
})
class TestAliasAddress(harness.TestCase):

    def test_invalid_init(self):
        with self.assertRaises(TypeError):
            self.type(self.data['type'], None)

    def test_properties(self):
        self.assertEqual(self.model.address, self.data['value'])


@harness.model_test_case({
    'type': models.Alias,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'type': models.AliasType.NONE,
        'value': None,
    },
    'eq': False,
    'dto': {
        'type': 0,
    },
})
class TestAliasEmpty(harness.TestCase):
    pass


@harness.model_test_case({
    'type': models.Alias,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'type': models.AliasType.MOSAIC_ID,
        'value': models.MosaicId(5),
    },
    'dto': {
        'type': 1,
        'mosaicId': [5, 0],
    },
})
class TestAliasMosaicId(harness.TestCase):

    def test_properties(self):
        self.assertEqual(self.model.mosaic_id, self.data['value'])

    def test_eq(self):
        self.assertEqual(self.model, self.model.replace())
        self.assertNotEqual(self.model, self.type())
        self.assertNotEqual(self.model, None)


@harness.enum_test_case({
    'type': models.AliasActionType,
    'enums': [
        models.AliasActionType.LINK,
        models.AliasActionType.UNLINK,
    ],
    'values': [
        0,
        1,
    ],
    'descriptions': [
        'Link an alias.',
        'Unlink an alias.',
    ],
    'dto': [
        0,
        1,
    ],
    'catbuffer': [
        b'\x00',
        b'\x01',
    ],
})
class TestAliasActionType(harness.TestCase):
    pass


@harness.enum_test_case({
    'type': models.AliasType,
    'enums': [
        models.AliasType.NONE,
        models.AliasType.MOSAIC_ID,
        models.AliasType.ADDRESS,
    ],
    'values': [
        0,
        1,
        2,
    ],
    'descriptions': [
        'No alias.',
        'Mosaic ID alias.',
        'Address alias.',
    ],
    'dto': [
        0,
        1,
        2,
    ],
    'catbuffer': [
        b'\x00',
        b'\x01',
        b'\x02',
    ],
})
class TestAliasType(harness.TestCase):
    pass


@harness.model_test_case({
    'type': models.EmptyAlias,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'value': None,
    },
    'eq': False,
    'dto': {
        'type': 0,
    },
    'extras': {
        'address': models.Address('SD5DT3CH4BLABL5HIMEKP2TAPUKF4NY3L5HRIR54'),
        'mosaic_id': models.MosaicId(5),
    },
})
class TestEmptyAlias(harness.TestCase):

    def test_rich_eq(self):
        empty_alias = models.Alias()
        address_alias = models.Alias(models.AliasType.ADDRESS, self.extras['address'])
        mosaic_alias = models.Alias(models.AliasType.MOSAIC_ID, self.extras['mosaic_id'])

        self.assertEqual(self.model, empty_alias)
        self.assertNotEqual(self.model, address_alias)
        self.assertNotEqual(self.model, mosaic_alias)

    def test_properties(self):
        with self.assertRaises(ValueError):
            self.model.address
        with self.assertRaises(ValueError):
            self.model.mosaic_id

    def test_from_invalid_dto(self):
        with self.assertRaises(ValueError):
            self.type.create_from_dto({'type': 1, 'mosaicId': [5, 0]})


@harness.model_test_case({
    'type': models.MosaicAlias,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'value': models.MosaicId(5),
    },
    'dto': {
        'type': 1,
        'mosaicId': [5, 0],
    },
    'extras': {
        'address': models.Address('SD5DT3CH4BLABL5HIMEKP2TAPUKF4NY3L5HRIR54'),
    },
})
class TestMosaicAlias(harness.TestCase):

    def test_rich_eq(self):
        empty_alias = models.Alias()
        address_alias = models.Alias(models.AliasType.ADDRESS, self.extras['address'])
        mosaic_alias = models.Alias(models.AliasType.MOSAIC_ID, self.data['value'])

        self.assertNotEqual(self.model, empty_alias)
        self.assertNotEqual(self.model, address_alias)
        self.assertEqual(self.model, mosaic_alias)

    def test_properties(self):
        with self.assertRaises(ValueError):
            self.model.address

    def test_from_invalid_dto(self):
        with self.assertRaises(ValueError):
            self.type.create_from_dto({'type': 0})


@harness.model_test_case({
    'type': models.NamespaceId,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'id': 5,
    },
})
class TestNamespaceId(harness.TestCase):

    def test_init(self):
        self.assertEqual(self.type(5).id, 5)
        self.assertEqual(self.type("").id, 0)
        self.assertEqual(self.type("sample").id, 0x88B64C3BE2F47144)
        self.assertEqual(self.type("sample.sub").id, 0xFA9429715A71ACC9)
        self.assertEqual(self.type("sample.sub.full").id, 0x8BC7011B0B344C54)

    def test_encoded(self):
        def encode(value):
            return self.type(value).encoded

        self.assertEqual('0500000000000000', self.model.encoded)
        self.assertEqual('4bfa5f372d55b384', encode(0x84b3552d375ffa4b))
        self.assertEqual('08a12f89ee5a49f8', encode(0xf8495aee892fa108))
        self.assertEqual('1f810565e8f4aeab', encode(0xabaef4e86505811f))
        self.assertEqual('552d1c0a2bc9b8ae', encode(0xaeb8c92b0a1c2d55))
        self.assertEqual('bfca1440d49ae090', encode(0x90e09ad44014cabf))
        self.assertEqual('ccf10b96814211ab', encode(0xab114281960bf1cc))

    def test_create_from_encoded(self):
        def from_encoded(value):
            return self.type.create_from_encoded(value)

        def test_from_encoded(value):
            self.assertEqual(from_encoded(value).encoded, value)

        test_from_encoded('4bfa5f372d55b384')
        test_from_encoded('08a12f89ee5a49f8')
        test_from_encoded('1f810565e8f4aeab')
        test_from_encoded('552d1c0a2bc9b8ae')
        test_from_encoded('bfca1440d49ae090')
        test_from_encoded('ccf10b96814211ab')

    def test_int(self):
        self.assertEqual(int(self.model), 5)

    def test_index(self):
        self.assertEqual(hex(self.model), "0x5")

    def test_format(self):
        value = self.type(5)
        self.assertEqual(f'{value:x}', '5')
        self.assertEqual(f'{value:X}', '5')

        value = self.type(13)
        self.assertEqual(f'{value:x}', 'd')
        self.assertEqual(f'{value:X}', 'D')

    def test_invalid_id(self):
        with self.assertRaises(TypeError):
            self.type(b'')


@harness.model_test_case({
    'type': models.NamespaceInfo,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'active': True,
        'index': 0,
        'meta_id': '5c7c07005cc1fe000176fa2b',
        'type': models.NamespaceType.ROOT_NAMESPACE,
        'depth': 1,
        'levels': [
            models.NamespaceId(0x84b3552d375ffa4b),
        ],
        'parent_id': models.NamespaceId(0),
        'owner': models.PublicAccount.create_from_public_key('7a562888c7ae1e082579951d6d93bf931de979360acca4c4085d754e5e122808', models.NetworkType.MIJIN_TEST),
        'start_height': 1,
        'end_height': 0xffffffffffffffff,
        'alias': models.Alias(),
    },
    'dto': {
        'meta': {
            'active': True,
            'index': 0,
            'id': '5c7c07005cc1fe000176fa2b'
        },
        'namespace': {
            'type': 0,
            'depth': 1,
            'level0': [929036875, 2226345261],
            'parentId': [0, 0],
            'owner': '7a562888c7ae1e082579951d6d93bf931de979360acca4c4085d754e5e122808',
            'ownerAddress': '90f6c07a4cf9ad1bf0644d419218b72fcdf1efcc07a6c9202c',
            'startHeight': [1, 0],
            'endHeight': [4294967295, 4294967295]
        }
    },
    'extras': {
        'parent': models.NamespaceId(0x88B64C3BE2F47144),
        'child': models.NamespaceId(0xFA9429715A71ACC9),
        'address': models.Address('SD5DT3CH4BLABL5HIMEKP2TAPUKF4NY3L5HRIR54'),
    }
})
class TestNamespaceInfo(harness.TestCase):

    def test_properties(self):
        self.assertEqual(self.model.id, models.NamespaceId(0x84b3552d375ffa4b))
        self.assertTrue(self.model.is_root())
        self.assertFalse(self.model.is_subnamespace())
        self.assertFalse(self.model.has_alias())

        with self.assertRaises(ValueError):
            self.model.parent_namespace_id()

        sub_namespace = self.model.replace(
            type=models.NamespaceType.SUB_NAMESPACE,
            parent_id=self.extras['parent'],
            levels=[self.extras['child']],
        )
        self.assertEqual(sub_namespace.parent_namespace_id(), self.extras['parent'])

    def test_alias_to_dto(self):
        model = self.model.replace(alias=models.AddressAlias(self.extras['address']))
        dto = model.to_dto(self.network_type)
        self.assertEqual(dto['namespace'].pop('alias'), {
            'type': 2,
            'address': '90fa39ec47e05600afa74308a7ea607d145e371b5f4f1447bc',
        })
        self.assertEqual(dto, self.dto)


@harness.model_test_case({
    'type': models.NamespaceName,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'namespace_id': models.NamespaceId(0x88b64c3be2f47144),
        'name': 'sample',
    },
    'dto': {
        'namespaceId': [0xe2f47144, 0x88b64c3b],
        'name': 'sample',
    },
})
class TestNamespaceNameDepth1(harness.TestCase):

    def test_create_from_name(self):
        self.assertEqual(self.model, self.type.create_from_name("sample"))
        self.assertTrue(self.model.is_valid())

    def test_valid(self):
        def create(name):
            # Create_from_name will detect the name is invalid, just do this.
            return self.type(models.NamespaceId(0), name)

        self.assertTrue(self.model.is_valid())
        self.assertFalse(create('').is_valid())
        self.assertTrue(create('lorem-ipsum-dolor-sit').is_valid())
        self.assertTrue(create('lorem_ipsum_dolor_sit').is_valid())
        self.assertFalse(create('Lorem_Ipsum_Dolor_Sit').is_valid())
        self.assertFalse(create('lorem ipsum dolor sit').is_valid())
        self.assertFalse(create('lorem_ipsum_dolor_sit_amet_consectetur_adipiscing_elit_sed_non_risus_suspendisse_lectus_tortor').is_valid())


@harness.model_test_case({
    'type': models.NamespaceName,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'namespace_id': models.NamespaceId(0xfa9429715a71acc9),
        'name': 'sub',
        'parent_id': models.NamespaceId(0x88b64c3be2f47144),
    },
    'dto': {
        'namespaceId': [0x5a71acc9, 0xfa942971],
        'name': 'sub',
        'parentId': [0xe2f47144, 0x88b64c3b],
    },
})
class TestNamespaceNameDepth2(harness.TestCase):

    def test_create_from_name(self):
        self.assertEqual(self.model, self.type.create_from_name("sample.sub"))
        self.assertTrue(self.model.is_valid())


@harness.model_test_case({
    'type': models.NamespaceName,
    'network_type': models.NetworkType.MIJIN_TEST,
    'data': {
        'namespace_id': models.NamespaceId(0x8bc7011b0b344c54),
        'name': 'full',
        'parent_id': models.NamespaceId(0xfa9429715a71acc9),
    },
    'dto': {
        'namespaceId': [0x0b344c54, 0x8bc7011b],
        'name': 'full',
        'parentId': [0x5a71acc9, 0xfa942971],
    },
})
class TestNamespaceNameDepth3(harness.TestCase):

    def test_create_from_name(self):
        self.assertEqual(self.model, self.type.create_from_name("sample.sub.full"))
        self.assertTrue(self.model.is_valid())


@harness.enum_test_case({
    'type': models.NamespaceType,
    'enums': [
        models.NamespaceType.ROOT_NAMESPACE,
        models.NamespaceType.SUB_NAMESPACE,
    ],
    'values': [
        0,
        1,
    ],
    'descriptions': [
        'Root namespace.',
        'Sub namespace.',
    ],
    'dto': [
        0,
        1,
    ],
    'catbuffer': [
        b'\x00',
        b'\x01',
    ],
})
class TestNamespaceType(harness.TestCase):
    pass



"""
    statements
    =================

    Describes statements.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from ..blockchain.network_type import OptionalNetworkType
from .transaction_statement import TransactionStatement
from .resolution_statement import ResolutionStatement
from ... import util


__all__ = ['Statements']


@util.inherit_doc
@util.dataclass(frozen=True)
class Statements(util.DTO):
    """
    Statements information.

    DTO Format:
        .. code-block:: yaml

            StatementsDTO:
                transactionStatements: TransactionStatementDTO[]
                addressResolutionStatements: ResolutionStatementDTO[]
                mosaicResolutionStatements: ResolutionStatementDTO[]
    """

    transaction_statements: typing.Sequence[TransactionStatement]
    address_resolution_statements: typing.Sequence[ResolutionStatement]
    mosaic_resolution_statements: typing.Sequence[ResolutionStatement]

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_l1 = {'transactionStatements', 'addressResolutionStatements', 'mosaicResolutionStatements'}
        return (
            # Level 1
            cls.validate_dto_required(data, required_l1)
            and cls.validate_dto_all(data, required_l1)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'transactionStatements': [i.to_dto(network_type) for i in self.transaction_statements],
            'addressResolutionStatements': [i.to_dto(network_type) for i in self.address_resolution_statements],
            'mosaicResolutionStatements': [i.to_dto(network_type) for i in self.mosaic_resolution_statements],
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        return cls(
            transaction_statements=[
                TransactionStatement.create_from_dto(i, network_type) for i in data['transactionStatements']
            ],
            address_resolution_statements=[
                ResolutionStatement.create_from_dto(i, network_type) for i in data['addressResolutionStatements']
            ],
            mosaic_resolution_statements=[
                ResolutionStatement.create_from_dto(i, network_type) for i in data['mosaicResolutionStatements']
            ]
        )



from xpxchain import models
from xpxchain import util
from tests import harness


class TestAccount(harness.TestCase):

    @harness.randomize(key={'fixed_length': 32})
    def test_create_from_private_key(self, key: bytes):
        network_type = models.NetworkType.MIJIN_TEST
        model = models.Account.create_from_private_key(key, network_type)
        pub = models.PublicAccount.create_from_public_key(model.public_key, network_type)
        self.assertTrue(model.address.is_valid())
        self.assertEqual(model.public_account, pub)

    @harness.randomize(key={'fixed_length': 33})
    def test_create_from_invalid_private_key(self, key: bytes):
        network_type = models.NetworkType.MIJIN_TEST
        with self.assertRaises(ValueError):
            models.Account.create_from_private_key(key, network_type)


class TestAddress(harness.TestCase):

    @harness.randomize(address={'pattern': r'[MSXVmsxv][A-Za-z2-7]{39}'})
    def test_mostly_valid(self, address: str):
        model = models.Address(address)
        self.assertEqual(model.address.lower(), address.lower())
        self.assertFalse(model.is_valid())

    @harness.randomize(address={'pattern': r'[A-LN-RT-UY-Za-ln-rt-uy-z][A-Za-z2-7]{39}'})
    def test_invalid_identifier(self, address: str):
        with self.assertRaises(KeyError):
            models.Address(address)

    @harness.randomize(encoded={'fixed_length': 20})
    def test_valid(self, encoded: bytes):
        encoded = b'\x90' + encoded
        checksum = util.hashlib.sha3_256(encoded).digest()[:4]
        model = models.Address.create_from_encoded(encoded + checksum)
        self.assertTrue(model.is_valid())

    @harness.randomize(key={'fixed_length': 32})
    def test_create_from_public_key(self, key: bytes):
        network_type = models.NetworkType.MIJIN_TEST
        model = models.Address.create_from_public_key(key, network_type)
        self.assertTrue(model.is_valid())

    @harness.randomize(key={'fixed_length': 33})
    def test_create_from_invalid_public_key(self, key: bytes):
        network_type = models.NetworkType.MIJIN_TEST
        with self.assertRaises(ValueError):
            models.Address.create_from_public_key(key, network_type)


class TestPublicAccount(harness.TestCase):

    @harness.randomize(key={'fixed_length': 32})
    def test_create_from_public_key(self, key: bytes):
        network_type = models.NetworkType.MIJIN_TEST
        model = models.PublicAccount.create_from_public_key(key, network_type)
        address = models.Address.create_from_public_key(key, network_type)
        self.assertTrue(model.address.is_valid())
        self.assertEqual(model.address, address)

    @harness.randomize(key={'fixed_length': 33})
    def test_create_from_invalid_public_key(self, key: bytes):
        network_type = models.NetworkType.MIJIN_TEST
        with self.assertRaises(ValueError):
            models.PublicAccount.create_from_public_key(key, network_type)



import binascii

from Crypto.Hash import RIPEMD160 as ripemd160_c
from xpxchain.util import hashlib
from tests import harness
from .helper import *


class TestRipemd160(harness.TestCase):

    def setUp(self):
        self.full_hexdigest = '8476ee4631b9b30ac2754b0ee0c47e161d3f724c'

    def test_init(self):
        actual = hexdigest(hashlib.ripemd160(), b'Hello World!')
        expected = hexdigest(ripemd160_c.new(), b'Hello World!')
        self.assertEqual(actual, expected)

    def test_properties(self):
        hasher = hashlib.ripemd160()
        self.assertEqual(hasher.digest_size, 20)

    def test_update(self):
        hasher = hashlib.ripemd160()
        hasher.update(b'Hello')
        hasher.update(b' ')
        hasher.update(b'World!')
        self.assertEqual(hasher.hexdigest(), self.full_hexdigest)

    def test_digest(self):
        self.assertEqual(digest(hashlib.ripemd160(), b'Hello World!'), binascii.unhexlify(self.full_hexdigest))

    def test_hexdigest(self):
        self.assertEqual(hexdigest(hashlib.ripemd160(), b'Hello World!'), self.full_hexdigest)

    def test_bytes(self):
        for test in STRING_TESTS:
            actual = hexdigest(hashlib.ripemd160(), test)
            expected = hexdigest(ripemd160_c.new(), test)
            self.assertEqual(actual, expected)



"""
    stdint
    ======

    Although Python has native support for arbitrary-precision integers,
    Javascript by default uses 64-bit floats as the only numeric type,
    signifying they cannot store more than 53 integral bits.

    Therefore, in Javascript, 64-bit integers are stored as an array
    of 2 numbers. Likewise, 128-bit integers are stored as an array of 4
    numbers. These functions the conversion of native Python integers
    to and from a Javascript-like notation, to simplify integration with
    data transfer objects.

    This also provides routines to convert to and from fixed-width
    integers in both catbuffer and data-transfer objects, as well
    as extract high- and low-bit patterns from the types.

    The module is named after <stdint.h>, which describes fixed-width
    (standard) integers in C, even though it has no relationship
    in terms of functionality.
"""

from __future__ import annotations
import typing

__all__ = [
    # DTO Types
    'I8DTOType',
    'U8DTOType',
    'U16DTOType',
    'U32DTOType',
    'U64DTOType',
    'U128DTOType',

    # Byte sizes
    'I8_BYTES',
    'U8_BYTES',
    'U16_BYTES',
    'U32_BYTES',
    'U64_BYTES',
    'U128_BYTES',

    # I8
    # 'i8_high',
    # 'i8_low',
    'i8_iter_from_catbuffer',
    'i8_iter_from_dto',
    'i8_iter_to_catbuffer',
    'i8_iter_to_dto',
    'i8_from_catbuffer',
    'i8_from_dto',
    'i8_to_catbuffer',
    'i8_to_dto',

    # U8
    'u8_high',
    'u8_low',
    'u8_iter_from_catbuffer',
    'u8_iter_from_dto',
    'u8_iter_to_catbuffer',
    'u8_iter_to_dto',
    'u8_from_catbuffer',
    'u8_from_dto',
    'u8_to_catbuffer',
    'u8_to_dto',

    # U16
    'u16_high',
    'u16_low',
    'u16_iter_from_catbuffer',
    'u16_iter_from_dto',
    'u16_iter_to_catbuffer',
    'u16_iter_to_dto',
    'u16_from_catbuffer',
    'u16_from_dto',
    'u16_to_catbuffer',
    'u16_to_dto',

    # U32
    'u32_high',
    'u32_low',
    'u32_iter_from_catbuffer',
    'u32_iter_from_dto',
    'u32_iter_to_catbuffer',
    'u32_iter_to_dto',
    'u32_from_catbuffer',
    'u32_from_dto',
    'u32_to_catbuffer',
    'u32_to_dto',

    # U64
    'u64_high',
    'u64_low',
    'u64_iter_from_catbuffer',
    'u64_iter_from_dto',
    'u64_iter_to_catbuffer',
    'u64_iter_to_dto',
    'u64_from_catbuffer',
    'u64_from_dto',
    'u64_to_catbuffer',
    'u64_to_dto',

    # U128
    'u128_high',
    'u128_low',
    'u128_iter_from_catbuffer',
    'u128_iter_from_dto',
    'u128_iter_to_catbuffer',
    'u128_iter_to_dto',
    'u128_from_catbuffer',
    'u128_from_dto',
    'u128_to_catbuffer',
    'u128_to_dto',
]

U4_BITS = 4
U8_BITS = 8
U16_BITS = 16
U32_BITS = 32
U64_BITS = 64
U128_BITS = 128

U8_BYTES = U8_BITS // 8
U16_BYTES = U16_BITS // 8
U32_BYTES = U32_BITS // 8
U64_BYTES = U64_BITS // 8
U128_BYTES = U128_BITS // 8

U4_MAX = 0xF
U8_MAX = 0xFF
U16_MAX = 0xFFFF
U32_MAX = 0xFFFFFFFF
U64_MAX = 0xFFFFFFFFFFFFFFFF
U128_MAX = 0xFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFFF

I8_BITS = 8

I8_BYTES = I8_BITS // 8

I8_MAX = 0x7F

I8_MIN = -0x80

U8DTOType = int
U16DTOType = int
U32DTOType = int
U64DTOType = typing.Sequence[U32DTOType]
U128DTOType = typing.Sequence[U64DTOType]
YieldIntType = typing.Generator[int, None, None]
YieldBytesType = typing.Generator[bytes, None, None]

I8DTOType = int

# HELPERS


def check_overflow(within_range: bool):
    """Raise exception if overflow."""
    if not within_range:
        raise OverflowError


def high(max: int, bits: int, mask: int) -> typing.Callable[[int], int]:
    def wrapper(value: int) -> int:
        check_overflow(0 <= value <= max)
        return (value >> bits) & mask

    wrapper.__name__ = f'u{2*bits}_high'
    wrapper.__qualname__ = wrapper.__name__
    wrapper.__doc__ = f'Get high {bits} from {2*bits}-bit integer.'
    wrapper.__module__ = __name__
    return wrapper


def low(max: int, bits: int, mask: int) -> typing.Callable[[int], int]:
    def wrapper(value: int) -> int:
        check_overflow(0 <= value <= max)
        return value & mask

    wrapper.__name__ = f'u{2*bits}_low'
    wrapper.__qualname__ = wrapper.__name__
    wrapper.__doc__ = f'Get low {bits} from {2*bits}-bit integer.'
    wrapper.__module__ = __name__
    return wrapper


def to_catbuffer_impl(size: int, signed: bool = False) -> typing.Callable[[int], bytes]:
    def wrapper(value: int, signed: bool = False) -> bytes:
        return value.to_bytes(size, 'little', signed=signed)
    return wrapper


def to_catbuffer(bits: int, signed: bool = False) -> typing.Callable[[int], bytes]:
    cb = to_catbuffer_impl(bits // 8, signed=signed)

    def wrapper(value: int, signed: bool = False) -> bytes:
        return cb(value)

    wrapper.__name__ = f'u{bits}_to_catbuffer'
    wrapper.__qualname__ = wrapper.__name__
    wrapper.__doc__ = f'Convert {bits}-bit integer to catbuffer.'
    wrapper.__module__ = __name__
    return wrapper


def iter_to_catbuffer(bits: int, signed: bool = False):
    cb = to_catbuffer_impl(bits // 8, signed=signed)

    def wrapper(iterable, signed: bool = False):
        for value in iterable:
            yield cb(value)

    wrapper.__name__ = f'u{bits}_iter_to_catbuffer'
    wrapper.__qualname__ = wrapper.__name__
    wrapper.__doc__ = f'Iteratively convert {bits}-bit integers to catbuffer.'
    wrapper.__module__ = __name__
    return wrapper


def from_catbuffer_impl(bits: int, signed: bool = False) -> typing.Callable[[bytes], int]:
    def wrapper(catbuffer: bytes, signed: bool = False) -> int:
        return int.from_bytes(catbuffer, 'little', signed=signed)
    return wrapper


def from_catbuffer(bits: int, signed: bool = False) -> typing.Callable[[bytes], int]:
    size = bits // 8
    cb = from_catbuffer_impl(size, signed=signed)

    def wrapper(catbuffer: bytes, signed: bool = False) -> int:
        if len(catbuffer) > size:
            raise OverflowError('bytes too big to convert')
        return cb(catbuffer)

    wrapper.__name__ = f'u{bits}_from_catbuffer'
    wrapper.__qualname__ = wrapper.__name__
    wrapper.__doc__ = f'Convert catbuffer to {bits}-bit integer.'
    wrapper.__module__ = __name__
    return wrapper


def iter_from_catbuffer(bits: int, signed: bool = False) -> typing.Callable[[bytes], YieldIntType]:
    size = bits // 8
    cb = from_catbuffer_impl(size, signed=signed)

    def wrapper(catbuffer: bytes, signed: bool = False) -> YieldIntType:
        length = len(catbuffer)
        if length % size != 0:
            raise ValueError(f'iter from_catbuffer requires multiple of {size}.')

        for i in range(0, length, size):
            start = i
            stop = start + size
            yield cb(catbuffer[start:stop])

    wrapper.__name__ = f'u{bits}_iter_from_catbuffer'
    wrapper.__qualname__ = wrapper.__name__
    wrapper.__doc__ = f'Iteratively convert catbuffer to {bits}-bit integers.'
    wrapper.__module__ = __name__
    return wrapper


def iter_to_dto(bits: int, cb):
    def wrapper(iterable):
        for value in iterable:
            yield cb(value)

    wrapper.__name__ = f'u{bits}_iter_to_dto'
    wrapper.__qualname__ = wrapper.__name__
    wrapper.__doc__ = f'Iteratively convert {bits}-bit integers to DTO.'
    wrapper.__module__ = __name__
    return wrapper


def iter_from_dto(bits: int, cb):
    def wrapper(iterable):
        for value in iterable:
            yield cb(value)

    wrapper.__name__ = f'u{bits}_iter_from_dto'
    wrapper.__qualname__ = wrapper.__name__
    wrapper.__doc__ = f'Iteratively convert DTOs to {bits}-bit integers.'
    wrapper.__module__ = __name__
    return wrapper


# UINT8


def u8_to_dto(value: int) -> U8DTOType:
    """Convert 8-bit int to DTO."""

    check_overflow(0 <= value <= U8_MAX)
    return value


def u8_from_dto(dto: U8DTOType) -> int:
    """Convert DTO to 8-bit int."""

    check_overflow(0 <= dto <= U8_MAX)
    return dto


u8_high = high(U8_MAX, U4_BITS, U4_MAX)
u8_low = low(U8_MAX, U4_BITS, U4_MAX)
u8_to_catbuffer = to_catbuffer(U8_BITS)
u8_from_catbuffer = from_catbuffer(U8_BITS)
u8_iter_to_catbuffer = iter_to_catbuffer(U8_BITS)
u8_iter_from_catbuffer = iter_from_catbuffer(U8_BITS)
u8_iter_to_dto = iter_to_dto(U8_BITS, u8_to_dto)
u8_iter_from_dto = iter_from_dto(U8_BITS, u8_from_dto)


# INT8


def i8_to_dto(value: int) -> I8DTOType:
    """Convert 8-bit int to DTO."""

    check_overflow(I8_MIN <= value <= I8_MAX)
    return value


def i8_from_dto(dto: I8DTOType) -> int:
    """Convert DTO to 8-bit int."""

    check_overflow(I8_MIN <= dto <= I8_MAX)
    return dto


# i8_high = high(I8_MAX, I4_BITS, I4_MAX)
# i8_low = low(I8_MAX, I4_BITS, I4_MAX)
i8_to_catbuffer = to_catbuffer(I8_BITS, signed=True)
i8_from_catbuffer = from_catbuffer(I8_BITS, signed=True)
i8_iter_to_catbuffer = iter_to_catbuffer(I8_BITS, signed=True)
i8_iter_from_catbuffer = iter_from_catbuffer(I8_BITS, signed=True)
i8_iter_to_dto = iter_to_dto(I8_BITS, i8_to_dto)
i8_iter_from_dto = iter_from_dto(I8_BITS, i8_from_dto)


# UINT16


def u16_to_dto(value: int) -> U16DTOType:
    """Convert 16-bit int to DTO."""

    check_overflow(0 <= value <= U16_MAX)
    return value


def u16_from_dto(dto: U16DTOType) -> int:
    """Convert DTO to 16-bit int."""

    check_overflow(0 <= dto <= U16_MAX)
    return dto


u16_high = high(U16_MAX, U8_BITS, U8_MAX)
u16_low = low(U16_MAX, U8_BITS, U8_MAX)
u16_to_catbuffer = to_catbuffer(U16_BITS)
u16_from_catbuffer = from_catbuffer(U16_BITS)
u16_iter_to_catbuffer = iter_to_catbuffer(U16_BITS)
u16_iter_from_catbuffer = iter_from_catbuffer(U16_BITS)
u16_iter_to_dto = iter_to_dto(U16_BITS, u16_to_dto)
u16_iter_from_dto = iter_from_dto(U16_BITS, u16_from_dto)

# UINT32


def u32_to_dto(value: int) -> U32DTOType:
    """Convert 32-bit int to DTO."""

    check_overflow(0 <= value <= U32_MAX)
    return value


def u32_from_dto(dto: U32DTOType) -> int:
    """Convert DTO to 32-bit int."""

    check_overflow(0 <= dto <= U32_MAX)
    return dto


u32_high = high(U32_MAX, U16_BITS, U16_MAX)
u32_low = low(U32_MAX, U16_BITS, U16_MAX)
u32_to_catbuffer = to_catbuffer(U32_BITS)
u32_from_catbuffer = from_catbuffer(U32_BITS)
u32_iter_to_catbuffer = iter_to_catbuffer(U32_BITS)
u32_iter_from_catbuffer = iter_from_catbuffer(U32_BITS)
u32_iter_to_dto = iter_to_dto(U32_BITS, u32_to_dto)
u32_iter_from_dto = iter_from_dto(U32_BITS, u32_from_dto)

# UINT64


def u64_to_dto(value: int) -> U64DTOType:
    """Convert 64-bit int to DTO."""

    check_overflow(0 <= value <= U64_MAX)
    return [u64_low(value), u64_high(value)]


def u64_from_dto(dto: U64DTOType) -> int:
    """Convert DTO to 64-bit int."""

    if not (
        len(dto) == 2
        and dto[0] <= U32_MAX
        and dto[1] <= U32_MAX
    ):
        raise ArithmeticError

    return (dto[0]) | (dto[1] << U32_BITS)


u64_high = high(U64_MAX, U32_BITS, U32_MAX)
u64_low = low(U64_MAX, U32_BITS, U32_MAX)
u64_to_catbuffer = to_catbuffer(U64_BITS)
u64_from_catbuffer = from_catbuffer(U64_BITS)
u64_iter_to_catbuffer = iter_to_catbuffer(U64_BITS)
u64_iter_from_catbuffer = iter_from_catbuffer(U64_BITS)
u64_iter_to_dto = iter_to_dto(U64_BITS, u64_to_dto)
u64_iter_from_dto = iter_from_dto(U64_BITS, u64_from_dto)

# UINT128


def u128_to_dto(value: int) -> U128DTOType:
    """Convert 128-bit int to DTO."""

    check_overflow(0 <= value <= U128_MAX)
    low = u128_low(value)
    high = u128_high(value)
    return [u64_to_dto(low), u64_to_dto(high)]


def u128_from_dto(dto: U128DTOType) -> int:
    """Convert DTO to 128-bit int."""

    if len(dto) != 2:
        raise ArithmeticError

    low = u64_from_dto(dto[0])
    high = u64_from_dto(dto[1])
    return low | (high << U64_BITS)


u128_high = high(U128_MAX, U64_BITS, U64_MAX)
u128_low = low(U128_MAX, U64_BITS, U64_MAX)
u128_to_catbuffer = to_catbuffer(U128_BITS)
u128_from_catbuffer = from_catbuffer(U128_BITS)
u128_iter_to_catbuffer = iter_to_catbuffer(U128_BITS)
u128_iter_from_catbuffer = iter_from_catbuffer(U128_BITS)
u128_iter_to_dto = iter_to_dto(U128_BITS, u128_to_dto)
u128_iter_from_dto = iter_from_dto(U128_BITS, u128_from_dto)



"""
    asyncio
    =======

    Asynchronous utilities.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import asyncio
import typing

__all__ = [
    'LoopType',
    'OptionalLoopType',
    'get_event_loop',
    'get_running_loop',
]

LoopType = asyncio.AbstractEventLoop
OptionalLoopType = typing.Optional[LoopType]


def get_event_loop(loop: OptionalLoopType = None) -> LoopType:
    """Get event loop."""

    if loop is None:
        loop = asyncio.get_event_loop()
    return loop


def get_running_loop(loop: OptionalLoopType = None) -> LoopType:
    """Get running event loop."""

    loop = get_event_loop(loop)
    if not loop.is_running():
        raise RuntimeError("No running event loop.")
    return loop



"""
    transaction_statement
    =================

    Describes a transaction statement.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations

import typing
from ..blockchain.network_type import OptionalNetworkType
from .balance_change_receipt import BalanceChangeReceipt
from .balance_transfer_receipt import BalanceTransferReceipt
from .artifact_expiry_receipt import ArtifactExpiryReceipt
from .inflation_receipt import InflationReceipt
from .receipt import Receipt
from .source import Source
from ... import util


__all__ = ['TransactionStatement']

ReceiptValue = typing.Union[
    BalanceChangeReceipt,
    BalanceTransferReceipt,
    ArtifactExpiryReceipt,
    InflationReceipt,
]
ReceiptValueList = typing.Sequence[ReceiptValue]


@util.inherit_doc
@util.dataclass(frozen=True)
class TransactionStatement(util.DTO):
    """
    Transaction statement information.

    :param height:
    :param source:
    :param receipts: The array of receipts.

    DTO Format:
        .. code-block:: yaml

            TransactionStatementsDTO:
                height: UInt64DTO
                source: SourceDTO
                receipts: anyOf(
                    BalanceTransferReceiptDTO,
                    BalanceChangeReceiptDTO,
                    ArtifactExpiryReceiptDTO,
                    InflationReceiptDTO
                )[]
    """

    height: int
    source: Source
    receipts: typing.Sequence[Receipt]

    @classmethod
    def validate_dto(cls, data: dict) -> bool:
        """Validate the data-transfer object."""

        required_l1 = {'height', 'source', 'receipts'}
        return (
            # Level 1
            cls.validate_dto_required(data, required_l1)
            and cls.validate_dto_all(data, required_l1)
        )

    def to_dto(
        self,
        network_type: OptionalNetworkType = None,
    ) -> dict:
        return {
            'height': util.u64_to_dto(self.height),
            'source': self.source.to_dto(network_type),
            'receipts': [i.to_dto(network_type) for i in self.receipts]
        }

    @classmethod
    def create_from_dto(
        cls,
        data: dict,
        network_type: OptionalNetworkType = None,
    ):
        if not cls.validate_dto(data):
            raise ValueError('Invalid data-transfer object.')

        return cls(
            height=util.u64_from_dto(data['height']),
            source=Source.create_from_dto(data['source'], network_type),
            receipts=[Receipt.create_from_dto(i, network_type) for i in data['receipts']]
        )



"""
    format
    ======

    Class-bound helpers to simplify conversion between interchange formats.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import typing

from ... import util


class FormatBase(util.Object):
    """Utilities to simplify loading and saving to interchange formats."""

    def save(self, key, data, value) -> None:
        """Save value to interchange format by key."""
        raise util.AbstractMethodError

    def load(self, key, data):
        """Load value from interchange format by key."""
        raise util.AbstractMethodError

    def load_type(self, data):
        """Load transaction type."""
        return self.load('type', data)

    def find_receipt(self, type_map, data):
        """Find derived receipt class via the receipt type."""
        return type_map[self.load_type(data)]


# DTO HELPERS
@util.dataclass(frozen=True)
class DTOFormat(FormatBase):
    """Utilities to simplify loading and saving to catbuffer."""

    names: typing.Dict[str, str]

    def save(self, key, data, value) -> None:
        name = self.names[key]
        cb = SAVE_DTO[key]
        if value is not None:
            data[name] = cb(value)

    def load(self, key, data):
        name = self.names[key]
        cb = LOAD_DTO[key]
        value = data.get(name)
        if value is not None:
            return cb(value)


def save_version_dto(version):
    return version


def save_type_dto(data):
    return data


def load_version_dto(version):
    return version


def load_type_dto(data):
    return data


SAVE_DTO = {
    'version': save_version_dto,
    'type': save_type_dto,
}

LOAD_DTO = {
    'version': load_version_dto,
    'type': load_type_dto,
}



from xpxchain import client
from xpxchain import models
from tests import harness
from tests import responses


@harness.mocked_http_test_case({
    'clients': (client.BlockchainHTTP, client.AsyncBlockchainHTTP),
    'network_type': models.NetworkType.MIJIN_TEST,
    'tests': [
        {
            'name': 'test_get_block_by_height',
            'response': responses.BLOCK_INFO["Ok"],
            'params': [1],
            'method': 'get_block_by_height',
            'validation': [
                lambda x: (x.hash, '04628F15E89936912DCA6345D60E668F513D50E7FC6BAF3F1C723C92CFC4FED6'),
                lambda x: (x.generation_hash, '7CCDF81A60B0A03A3B30D715C5C7513916319C09215E22C67B0A81106FB21445'),
                lambda x: (x.total_fee, 0),
                lambda x: (x.num_transactions, 31),
                lambda x: (x.signature, "82D8C518BF32CA9A8478451F79B7EEC69F11E6022AF7230CE21F230A09E51A0CEB169454CB4AD4499BC9C0769B468AB5A6DD716D00C977F16BF63DD1B0653305"),
                lambda x: (x.signer.public_key, "A04335F99D9EE3787528A16C7A302F80D511E9CF71D97D95C2182E0EA75A1EF9"),
                lambda x: (x.network_type, models.NetworkType.MIJIN_TEST),
                lambda x: (x.version, 3),
                lambda x: (x.type, 32835),
                lambda x: (x.height, 1),
                lambda x: (x.timestamp, 0),
                lambda x: (x.fee_multiplier, 0),
                lambda x: (x.difficulty, 100000000000000),
                lambda x: (x.previous_block_hash, "0000000000000000000000000000000000000000000000000000000000000000"),
                lambda x: (x.block_transactions_hash, "36822DF1CF1827915A8CDDF8A47EBBF1D1C55666DB1E8AC942C447D3D0E6D494"),
                lambda x: (x.block_receipts_hash, "RMViDPTpgKd072qPgxmPVoabWcrmFs5v5EGS00GxiX0="),
                lambda x: (x.state_hash, "C61C28E078298D6A508A649637EAC7D9D1E5F3813856EF9B7F1579866ACB570E"),
                lambda x: (x.merkle_tree, []),
            ]
        },
        {
            'name': 'test_get_blocks_by_height_with_limit',
            'response': responses.BLOCKS_INFO["Ok"],
            'params': [1, 1],
            'method': 'get_blocks_by_height_with_limit',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (x[0].hash, '04628F15E89936912DCA6345D60E668F513D50E7FC6BAF3F1C723C92CFC4FED6'),
            ]
        },
        {
            'name': 'test_get_block_transactions',
            'response': responses.BLOCK_TRANSACTIONS["Ok"],
            'params': [1],
            'method': 'get_block_transactions',
            'validation': [
                lambda x: (len(x), 1),
                lambda x: (x[0].transaction_info.height, 1),
                lambda x: (x[0].transaction_info.hash, "0E814F8D301E90144F1F25384C05AC26EF9D45FACAD216590B876FBB09877E02"),
                lambda x: (x[0].transaction_info.merkle_component_hash, "0E814F8D301E90144F1F25384C05AC26EF9D45FACAD216590B876FBB09877E02"),
                lambda x: (x[0].signature, "C214F728EA91BB9947F843CB6FF59EB8D231DB652796D25D055E4D1D8E65BF67F8800B268257E9478418E1AA0D7C8DB451E82176437D21F11EB777384583230B"),
                lambda x: (x[0].signer.public_key, "A04335F99D9EE3787528A16C7A302F80D511E9CF71D97D95C2182E0EA75A1EF9"),
                lambda x: (x[0].version, 2),
                lambda x: (x[0].type, 16718),
                lambda x: (x[0].max_fee, 0),
                lambda x: (x[0].namespace_type, models.NamespaceType.ROOT_NAMESPACE),
                lambda x: (x[0].duration, 0),
                lambda x: (x[0].namespace_id, models.NamespaceId.create_from_hex('b1497f5fba651b4f')),
                lambda x: (x[0].namespace_name, 'cat'),
            ]
        },
        {
            'name': 'test_get_blockchain_height',
            'response': responses.CHAIN_HEIGHT["Ok"],
            'params': [],
            'method': 'get_blockchain_height',
            'validation': [
                lambda x: (x, 53577),
            ]
        },
        {
            'name': 'test_get_blockchain_score',
            'response': responses.CHAIN_SCORE["Ok"],
            'params': [],
            'method': 'get_blockchain_score',
            'validation': [
                lambda x: (isinstance(x, models.BlockchainScore), True),
                lambda x: (x.score, 0x791f69c466a3658),
                lambda x: (x.score_low, 0x791f69c466a3658),
                lambda x: (x.score_high, 0),
            ]
        },
        {
            'name': 'test_get_diagnostic_blocks_by_height_with_limit',
            'response': responses.DIAGNOSTIC_BLOCKS_INFO["Ok"],
            'params': [1, 5],
            'method': 'get_diagnostic_blocks_by_height_with_limit',
            'validation': [
                lambda x: (len(x), 5),
                lambda x: (x[0].hash, '22A2B807BF395CC9432DB4075B8A6D412EF7FD0892D266959D02D942C0F48576'),
                lambda x: (x[0].merkle_tree[0], '796159C97C02FB6F38BB7C07F9F367A5E7362090D4FDA34739979A6647E127D3'),
                lambda x: (x[1].hash, 'DF7539D4AC9F8B1DA736CB1F2BEFCD39521ECCDB927929307A1FAE0B1E60FB39'),
                lambda x: (x[2].hash, '9D004923FA3D31928BFB961E4D84E3A42A1C34F75DC3584EDB6E87520968FDDB'),
                lambda x: (x[3].hash, '5228EA55037FE6B5AFBB25B791194A12F9D7FA99CF3FAE1E6B99A147999B3FD5'),
                lambda x: (x[4].hash, '04628F15E89936912DCA6345D60E668F513D50E7FC6BAF3F1C723C92CFC4FED6'),
            ]
        },
        {
            'name': 'test_get_diagnostic_storage',
            'response': responses.DIAGNOSTIC_STORAGE["Ok"],
            'params': [],
            'method': 'get_diagnostic_storage',
            'validation': [
                lambda x: (isinstance(x, models.BlockchainStorageInfo), True),
                lambda x: (x.num_blocks, 53582),
                lambda x: (x.num_transactions, 25),
                lambda x: (x.num_accounts, 25),
            ]
        },
        {
            'name': 'test_get_merkle_by_hash_in_block',
            'response': responses.BLOCK_TRANSACTION_MERKLE["Ok"],
            'params': [1, '6C274716520ABCDCE0BB799CF6F82B64EBC9B954DBE3B14FD5767CD1A89970BE'],
            'method': 'get_merkle_by_hash_in_block',
            'validation': [
                lambda x: (isinstance(x, models.MerkleProofInfo), True),
                lambda x: (len(x.merkle_path), 6),
                lambda x: (isinstance(x.merkle_path[0], models.MerklePathItem), True),
                lambda x: (x.merkle_path[0].position, 2),
                lambda x: (x.merkle_path[0].hash, 'D41D8437B13EFEBCD9077A325B0604E4FF6691FDE13D6D46B318508EC09B8DA7'),
            ]
        },
        {
            'name': 'test_get_block_receipts',
            'response': responses.BLOCK_RECEIPTS["Ok"],
            'params': [25],
            'method': 'get_block_receipts',
            'validation': [
                lambda x: (isinstance(x, models.Statements), True),
                lambda x: (len(x.transaction_statements), 1),
                lambda x: (len(x.address_resolution_statements), 0),  # TODO: Add statements
                lambda x: (len(x.mosaic_resolution_statements), 0),  # TODO: Add statements
                lambda x: (x.transaction_statements[0].height, 25),
                lambda x: (isinstance(x.transaction_statements[0].source, models.Source), True),
                lambda x: (x.transaction_statements[0].source.primary_id, 0),
                lambda x: (x.transaction_statements[0].source.secondary_id, 0),
                lambda x: (len(x.transaction_statements[0].receipts), 1),
                lambda x: (isinstance(x.transaction_statements[0].receipts[0], models.Receipt), True),
                lambda x: (x.transaction_statements[0].receipts[0].type, models.ReceiptType.VALIDATE_FEE),
                lambda x: (x.transaction_statements[0].receipts[0].version, 1),
                lambda x: (isinstance(x.transaction_statements[0].receipts[0].account, models.PublicAccount), True),
                lambda x: (x.transaction_statements[0].receipts[0].account.public_key, '346E56F77F07B19D48B3AEF969EDB01F68A5AC9FAF8E5E007577D71BA66385FB'),
                lambda x: (isinstance(x.transaction_statements[0].receipts[0].mosaic, models.Mosaic), True),
                lambda x: (x.transaction_statements[0].receipts[0].mosaic.id, models.MosaicId(992621222383397347)),
                lambda x: (x.transaction_statements[0].receipts[0].mosaic.amount, 0),
            ]
        },
        {
            'name': 'test_get_diagnostic_server',
            'response': responses.DIAGNOSTIC_SERVER["Ok"],
            'params': [],
            'method': 'get_diagnostic_server',
            'validation': [
                lambda x: (isinstance(x, models.BlockchainServerInfo), True),
                lambda x: (x.rest_version, '1.0.16'),
                lambda x: (x.sdk_version, '0.7.16'),
            ]
        }
    ],
})
class TestBlockchainHTTP(harness.TestCase):
    pass



"""
    deadline
    ========

    Deadline of a transaction.

    License
    -------

    Copyright 2019 NEM

    Licensed under the Apache License, Version 2.0 (the "License");
    you may not use this file except in compliance with the License.
    You may obtain a copy of the License at

        http://www.apache.org/licenses/LICENSE-2.0

    Unless required by applicable law or agreed to in writing, software
    distributed under the License is distributed on an "AS IS" BASIS,
    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
    See the License for the specific language governing permissions and
    limitations under the License.
"""

from __future__ import annotations
import datetime
import enum

from ... import util


__all__ = [
    'ChronoUnit',
    'Deadline',
    'TIMESTAMP_NEMESIS_BLOCK',
]


@util.inherit_doc
class ChronoUnit(util.EnumMixin, enum.IntEnum):
    """Enumerations for time units."""

    MICROSECONDS = 0
    MILLISECONDS = 1
    SECONDS = 2
    MINUTES = 3
    HOURS = 4

    def description(self) -> str:
        return DESCRIPTION[self]

    def to_timedelta(self, value: int) -> datetime.timedelta:
        """Get timedelta as multiple of units."""

        kw = KEYWORDS[self]
        return datetime.timedelta(**{kw: value})


DESCRIPTION = {
    ChronoUnit.MICROSECONDS: "Microseconds.",
    ChronoUnit.MILLISECONDS: "Milliseconds.",
    ChronoUnit.SECONDS: "Seconds.",
    ChronoUnit.MINUTES: "Minutes.",
    ChronoUnit.HOURS: "Hours.",
}

KEYWORDS = {
    ChronoUnit.MICROSECONDS: "microseconds",
    ChronoUnit.MILLISECONDS: "milliseconds",
    ChronoUnit.SECONDS: "seconds",
    ChronoUnit.MINUTES: "minutes",
    ChronoUnit.HOURS: "hours",
}

TIMESTAMP_NEMESIS_BLOCK: int = 1459468800000


@util.inherit_doc
@util.dataclass(frozen=True)
class Deadline(util.Object):
    """
    Deadline of a transaction.

    The number of seconds since the creation of the nemesis block.
    The number of seconds stored in the class are in local time, and
    any conversions to or from data-transfer objects will implicitly
    convert the data to UTC.

    :param deadline: Datetime of deadline in local time.
    """

    deadline: datetime.datetime

    @classmethod
    def create(
        cls,
        deadline: int = 2,
        unit: ChronoUnit = ChronoUnit.HOURS
    ):
        """
        Create deadline relative to current local time.

        :param deadline: Time for deadline after current time.
        :param chrono_unit: Unit for created deadline.
        """

        delta = unit.to_timedelta(deadline)
        day = datetime.timedelta(days=1).total_seconds()
        now = datetime.datetime.now()
        delta_seconds = delta.total_seconds()
        if delta_seconds <= 0:
            raise ValueError(f"Deadline should be greater than 0, got {delta_seconds}")
        elif delta_seconds >= day:
            raise ValueError(f"Deadline should be less than 1 day, got {delta_seconds}")

        return cls(now + delta)

    def to_timestamp(self) -> int:
        """Export deadline to UTC timestamp."""

        utc = self.deadline.replace(tzinfo=datetime.timezone.utc)
        return int(utc.timestamp() * 1000) - TIMESTAMP_NEMESIS_BLOCK

    @classmethod
    def create_from_timestamp(cls, timestamp: int):
        """
        Create deadline from UTC timestamp.

        :param timestamp: Timestamp in UTC timezone.
        """
        timestamp = (timestamp + TIMESTAMP_NEMESIS_BLOCK) / 1000
        utc = datetime.datetime.fromtimestamp(timestamp, datetime.timezone.utc)
        local = utc.replace(tzinfo=None)
        return cls(local)

    @classmethod
    def create_fake(cls, timestamp: int):
        """
        Create deadline from UTC timestamp.

        :param timestamp: Timestamp in UTC timezone.
        """
        ts = timestamp + TIMESTAMP_NEMESIS_BLOCK / 1000
        utc = datetime.datetime.fromtimestamp(ts, datetime.timezone.utc)
        local = utc.replace(tzinfo=None)
        return cls(local)


# Private
# Data-transfer object for the nemesis block timestamp.
TIMESTAMP_NEMESIS_BLOCK_DTO = util.u64_to_dto(TIMESTAMP_NEMESIS_BLOCK)
